##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Functions to generate matchup table from SUMO network XML file. """

import xml.etree.ElementTree as ET
import pandas as pd
from pyproj import Proj, Transformer
import math
from openpyxl import Workbook
from openpyxl.styles import Alignment
import copy


def get_net_edges(path_net: str) -> pd.DataFrame:
    """ Extract edges from a SUMO network XML file and return them as a DataFrame."""
    # Load the network XML file
    tree = ET.parse(path_net)
    root = tree.getroot()

    # Extract edges and store them in a DataFrame
    edges_data = []
    for edge in root.findall("edge"):
        edge_id = edge.get("id")
        from_junction = edge.get("from")
        to_junction = edge.get("to")
        edges_data.append(
            {"Edge ID": edge_id, "From": from_junction, "To": to_junction})
    df_edges = pd.DataFrame(edges_data)
    return df_edges


def get_net_connections(path_net: str) -> pd.DataFrame:
    """ Extract connections from a SUMO network XML file and return them as a DataFrame."""
    # Load the network XML file
    tree = ET.parse(path_net)
    root = tree.getroot()
    connections_data = []
    direction_mapping = {"s": "thru",
                         "l": "left", "L": "left",
                         "r": "right", "R": "right",
                         "t": "Uturn", "invalid": "invalid"}
    for connection in root.findall("connection"):
        from_edge = connection.get("from")
        to_edge = connection.get("to")
        direction_code = connection.get("dir")
        direction = direction_mapping.get(direction_code, None)
        connections_data.append({"FromEdge": from_edge, "ToEdge": to_edge, "Direction": direction})

    df_connections = pd.DataFrame(connections_data)
    return df_connections


def generate_matchup_table(df_matchup_table: pd.DataFrame, path_output: str = "MatchUp_Table.xlsx") -> bool:
    """ Generate a matchup table from the provided DataFrame and save it to an Excel file."""
    network_columns = ["JunctionID_OpenDrive", "Bearing", "Numbering", "FromRoadID_OpenDrive", "ToRoadID_OpenDrive", "Turn"]
    demand_columns = ["File_GridSmart", "Date_GridSmart", "IntersectionName_GridSmart", "Turn_GridSmart"]
    signal_columns = ["File_Synchro", "IntersectionID_Synchro", "Turn_Synchro"]
    other_columns = ["Need calibration?"]

    # df_matchup_table = path_net
    # Generate the matchup table
    # df_matchup_table = format_junction_bearing(path_net)

    wb = Workbook()
    ws = wb.active

    ws.append(["Network"] * len(network_columns) + ["Demand"] * len(demand_columns) +
              ["Signal"] * len(signal_columns) + [""] * len(other_columns))

    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=len(network_columns))
    ws.merge_cells(start_row=1, start_column=len(network_columns) + 1, end_row=1,
                   end_column=len(network_columns) + len(demand_columns))
    ws.merge_cells(start_row=1, start_column=len(network_columns) + len(demand_columns) + 1, end_row=1,
                   end_column=len(network_columns) + len(demand_columns) + len(signal_columns))

    ws.append(network_columns + demand_columns + signal_columns + other_columns)

    for row in df_matchup_table.itertuples(index=False):
        ws.append(list(row))

    current_start = 3  # Data starts at row 3
    for i in range(3, len(df_matchup_table) + 3):
        if (i == len(df_matchup_table) + 2 or ws[f"A{i}"].value != ws[f"A{i + 1}"].value):  # Check next row
            if current_start < i:  # Only merge if there are multiple same values
                ws.merge_cells(start_row=current_start, start_column=1, end_row=i, end_column=1)  # JunctionID_OpenDrive
                ws.merge_cells(start_row=current_start, start_column=7, end_row=i, end_column=7)  # File_GridSmart
                ws.merge_cells(start_row=current_start, start_column=8, end_row=i, end_column=8)  # Date_GridSmart
                ws.merge_cells(start_row=current_start, start_column=9, end_row=i, end_column=9)  # IntersectionName_GridSmart
                # ws.merge_cells(start_row=current_start, start_column=11, end_row=i, end_column=11)  # File_Synchro
                ws.merge_cells(start_row=current_start, start_column=12, end_row=i, end_column=12)  # IntersectionID_Synchro
                ws.merge_cells(start_row=current_start, start_column=14, end_row=i, end_column=14)  # Need calibration?
            current_start = i + 1

    if len(df_matchup_table) > 0:
        ws.merge_cells(start_row=3, start_column=11, end_row=len(
            df_matchup_table) + 2, end_column=11)

    # Center align merged cells
    for row in ws.iter_rows():
        for cell in row:
            cell.alignment = Alignment(horizontal="center", vertical="center")

    # Adjust column widths
    column_widths = {
        "A": 20,  # JunctionID_OpenDrive
        "B": 15,  # Bearing
        "C": 15,  # Numbering
        "D": 25,  # FromRoadID_OpenDrive
        "E": 25,  # ToRoadID_OpenDrive
        "F": 15,  # Turn
        "G": 20,  # File_GridSmart
        "H": 20,  # Date_GridSmart
        "I": 30,  # IntersectionName_GridSmart
        "J": 20,  # Turn_GridSmart
        "K": 20,  # File_Synchro
        "L": 25,  # IntersectionID_Synchro
        "M": 20,  # Turn_Synchro
        "N": 20   # Need calibration?
    }
    for col, width in column_widths.items():
        ws.column_dimensions[col].width = width
    wb.save(path_output)
    return True


def generate_junction_bearing(path_net: str) -> pd.DataFrame:
    """ Generate a DataFrame containing junction bearings from a SUMO network XML file."""
    tree = ET.parse(path_net)
    root = tree.getroot()

    # Extract edges and store them in a DataFrame
    df_edges = get_net_edges(path_net)

    # Extract correct UTM zone and netOffset
    location = root.find("location")
    if location is not None:
        net_offset_x, net_offset_y = [float(val) for val in location.get("netOffset", "0,0").split(",")]

        # Extract the UTM zone from the projection parameters
        proj_params = location.get("projParameter", "")
        utm_zone = None
        for param in proj_params.split():
            if "+zone=" in param:
                utm_zone = int(param.split("=")[1])
                break

    if utm_zone is None:
        raise ValueError("UTM zone could not be found in the XML file's projection parameters.")

    # Define the correct UTM projection with the extracted zone
    utm_proj = Proj(proj="utm", zone=utm_zone, ellps="WGS84", datum="WGS84", units="m", no_defs=True)
    wgs84_proj = Proj(proj="latlong", datum="WGS84")

    # Function to convert (x, y) to (lat, lon) using extracted net_offset and UTM zone
    def convert_coordinates(x, y):
        utm_x = x - net_offset_x
        utm_y = y - net_offset_y
        lon, lat = Transformer.from_proj(utm_proj, wgs84_proj, always_xy=True).transform(utm_x, utm_y)
        return lat, lon

    # Dictionary to store junctions with at least 3 approaches and their bearing coordinates
    junction_bearing = []

    for junction in root.findall("junction"):
        junction_id = junction.get("id")
        inc_lanes = junction.get("incLanes", "").split()

        # Extract unique edges from lane IDs
        edges = set(lane.rsplit("_", 1)[0] for lane in inc_lanes if lane)

        # Get entrance and exit counts for this junction
        # entrance_count = sum(df_edges["To"] == junction_id)
        exit_count = sum(df_edges["From"] == junction_id)

        # Check if the junction has at least 2 entrances or at least 2 exits
        # if entrance_count >= 2 or exit_count >= 2:
        if exit_count >= 2:
            for edge in root.findall("edge"):
                edge_id = edge.get("id")
                if edge_id in edges:
                    # Get the first lane of this edge
                    first_lane = edge.find("lane")
                    if first_lane is not None:
                        shape = first_lane.get("shape", "").split()

                        # Extract the last two points
                        if len(shape) >= 2:
                            last_point = shape[-1].split(",")[:2]  # Get x, y of the last point
                            second_last_point = shape[-2].split(",")[:2]  # Get x, y of the second last point

                            second_last_point = tuple(float(val) for val in second_last_point)
                            last_point = tuple(float(val) for val in last_point)

                            # Convert to lat, lon
                            lat1, lon1 = convert_coordinates(*second_last_point)
                            lat2, lon2 = convert_coordinates(*last_point)

                            # Calculate bearing
                            bearing = calculate_bearing(lat1, lon1, lat2, lon2)

                            # Store the approach with its coordinates
                            junction_bearing.append({
                                "Junction ID": junction_id,
                                "Approach Edge": edge_id,
                                "Second Last Point (x, y)": second_last_point,
                                "Last Point (x, y)": last_point,
                                "Second Last Point (lat, lon)": (lat1, lon1),
                                "Last Point (lat, lon)": (lat2, lon2),
                                "Degree": round(bearing, 2),
                                "Runway Bearing": int(round(bearing / 10.0) * 10 / 10)
                            })

    df_junction_bearing = pd.DataFrame(junction_bearing)
    return df_junction_bearing


def format_junction_bearing(path_net: str) -> pd.DataFrame:
    """ Format the junction bearing data from a SUMO network XML file into a matchup table."""

    # Generate the original junction bearing DataFrame
    junction_bearing = generate_junction_bearing(path_net)

    MatchupTable = copy.deepcopy(junction_bearing)
    MatchupTable.drop(columns=[
        "Second Last Point (x, y)", "Last Point (x, y)",
        "Second Last Point (lat, lon)", "Last Point (lat, lon)"
    ], inplace=True)
    MatchupTable.rename(columns={"Approach Edge": "FromEdge"}, inplace=True)
    MatchupTable.insert(MatchupTable.columns.get_loc("FromEdge") + 1, "ToEdge", None)
    MatchupTable.insert(MatchupTable.columns.get_loc("ToEdge") + 1, "Direction", None)

    df_connections = get_net_connections(path_net)
    MatchupTable = MatchupTable.merge(df_connections, on="FromEdge", how="left")
    MatchupTable.drop(columns=["ToEdge_x", "Direction_x"], inplace=True)
    MatchupTable.rename(columns={"ToEdge_y": "ToEdge", "Direction_y": "Direction"}, inplace=True)
    MatchupTable.drop_duplicates(subset=["Junction ID", "FromEdge", "ToEdge", "Direction"], inplace=True)
    MatchupTable.reset_index(drop=True, inplace=True)

    # MatchupTable["Junction ID Numeric"] = pd.to_numeric(MatchupTable["Junction ID"], errors='coerce')
    MatchupTable["Junction ID Numeric"] = MatchupTable["Junction ID"].astype(str)
    direction_order = {"right": 1, "thru": 2, "left": 3, "Uturn": 4}
    MatchupTable["Direction Order"] = MatchupTable["Direction"].map(direction_order)

    # Sort by "Junction ID", "Degree", and "Direction Order"
    # MatchupTable = MatchupTable.sort_values(
    #     by=["Junction ID Numeric", "Degree", "Direction Order"],
    #     ascending=[True, True, True],
    #     na_position='last'
    # )
    # MatchupTable.drop(columns=["Junction ID Numeric", "Direction Order"], inplace=True)
    # MatchupTable.reset_index(drop=True, inplace=True)

    MatchupTable['Degree'] = MatchupTable['Degree'].astype(float)
    # Shift Degree so that 337.5 becomes the new zero reference
    MatchupTable['Degree_shifted'] = (MatchupTable['Degree'] - 337.5) % 360
    # Sort by Junction ID, shifted Degree, and Direction Order
    MatchupTable = MatchupTable.sort_values(
        by=["Junction ID Numeric", "Degree_shifted", "Direction Order"],
        ascending=[True, True, True],
        na_position='last'
    )
    # Drop helper column
    MatchupTable.drop(columns=["Junction ID Numeric", "Direction Order", 'Degree_shifted'], inplace=True)
    MatchupTable.reset_index(drop=True, inplace=True)

    MatchupTable.rename(columns={
        "Junction ID": "JunctionID_OpenDrive",
        "FromEdge": "FromRoadID_OpenDrive",
        "Degree": "Bearing",
        "Runway Bearing": "Numbering",
        "ToEdge": "ToRoadID_OpenDrive",
        "Direction": "Turn"
    }, inplace=True)

    # remove "-" before the sumo edge id to be OpenDrive road id
    MatchupTable["FromRoadID_OpenDrive"] = MatchupTable["FromRoadID_OpenDrive"].astype(str).str.replace(r"^-", "", n=1, regex=True)
    MatchupTable["ToRoadID_OpenDrive"] = MatchupTable["ToRoadID_OpenDrive"].astype(str).str.replace(r"^-", "", n=1, regex=True)

    MatchupTable["File_GridSmart"] = None
    MatchupTable["Date_GridSmart"] = None
    MatchupTable["IntersectionName_GridSmart"] = None
    MatchupTable["Turn_GridSmart"] = None
    MatchupTable["File_Synchro"] = None
    MatchupTable["IntersectionID_Synchro"] = None
    MatchupTable["Turn_Synchro"] = None
    MatchupTable["Need calibration?"] = None

    MatchupTable = MatchupTable[[
        "JunctionID_OpenDrive", "Bearing", "Numbering", "FromRoadID_OpenDrive",
        "ToRoadID_OpenDrive", "Turn", "File_GridSmart", "Date_GridSmart",
        "IntersectionName_GridSmart", "Turn_GridSmart", "File_Synchro",
        "IntersectionID_Synchro", "Turn_Synchro", "Need calibration?"
    ]]
    return MatchupTable


# Function to calculate bearing
def calculate_bearing(lat1, lon1, lat2, lon2):
    """ Calculate the bearing between two geographical points."""
    delta_lon = math.radians(lon2 - lon1)
    lat1 = math.radians(lat1)
    lat2 = math.radians(lat2)
    x = math.sin(delta_lon) * math.cos(lat2)
    y = math.cos(lat1) * math.sin(lat2) - math.sin(lat1) * math.cos(lat2) * math.cos(delta_lon)
    initial_bearing = math.atan2(x, y)
    initial_bearing = math.degrees(initial_bearing)
    compass_bearing = (initial_bearing + 360) % 360
    return compass_bearing


# Function to match input road to best bearing
def match_best_bearing(junction_id, lat1, lon1, lat2, lon2, junction_bearing):
    """ Match the input road to the best bearing based on the junction bearing DataFrame."""
    if not isinstance(junction_id, str):
        junction_id = str(junction_id)

    input_bearing = calculate_bearing(lat1, lon1, lat2, lon2)
    # runout_bearing = int(round(input_bearing / 10.0) * 10 / 10)

    # Filter junction bearings for the given junction
    matching_junctions = junction_bearing[junction_bearing["Junction ID"] == junction_id].copy()

    if matching_junctions.empty:
        print(f"No matching junction found for {junction_id}.")
        return None, None, None

    # Find the best matching edge
    matching_junctions.loc[:, "Bearing Difference"] = abs(matching_junctions["Degree"] - input_bearing)
    best_match = matching_junctions.loc[matching_junctions["Bearing Difference"].idxmin()]

    # Output absolute difference instead of similarity
    abs_difference = best_match["Bearing Difference"]

    print(f"This road should be edge {best_match['Approach Edge']} "
          f"(runway bearing {int(round(best_match['Degree'] / 10.0) * 10 / 10)}) "
          f"of junction {junction_id}. The difference is {abs_difference:.2f}°.")

    return best_match['Approach Edge'], int(round(best_match['Degree'] / 10.0) * 10 / 10), abs_difference


# if __name__ == "__main__":
#     path_net = "./chatt.net.xml"
#     generate_matchup_table(path_net, "MatchupTable.xlsx")



'''
##############################################################
# Created Date: Wednesday, October 1st 2025
# Contact Info: luoxiangyong01@gmail.com
# Author/Copyright: Mr. Xiangyong Luo
##############################################################
'''


import sys
import pytest
from realtwin import check_python_version
from io import StringIO


def make_sys_version(version_str):
    """Helper to patch sys.version to a specific version string."""
    return version_str + " (default, Jan 1 2024, 00:00:00) [MSC v.1916 64 bit (AMD64)]"


@pytest.mark.parametrize(
    "sys_version,min_version,expected_tuple,should_raise",
    [
        ("3.10.0", "3.10", (3, 10, 0), False),
        ("3.11.2", "3.10", (3, 11, 2), False),
        ("3.9.9", "3.10", (3, 9, 9), True),
        ("3.10.1", "3.10", (3, 10, 1), False),
        ("3.10.0", "3.11", (3, 10, 0), True),
        ("3.11.0", "3.11", (3, 11, 0), False),
        ("3.12.0", "3.11", (3, 12, 0), False),
    ]
)
def test_check_python_version(monkeypatch, sys_version, min_version, expected_tuple, should_raise):
    # Patch sys.version
    monkeypatch.setattr(sys, "version", make_sys_version(sys_version))
    if should_raise:
        # The function prints a message and returns the tuple, does not raise
        # So we capture stdout to check the print
        captured = StringIO()
        monkeypatch.setattr(sys, "stdout", captured)
        result = check_python_version(min_version)
        assert result == expected_tuple
        output = captured.getvalue()
        assert f"pyufunc supports Python {min_version} or higher." in output
    else:
        result = check_python_version(min_version)
        assert result == expected_tuple


def test_check_python_version_default(monkeypatch):
    # Should use current sys.version and min_version="3.10"
    result = check_python_version()
    # Should return a tuple of ints
    assert isinstance(result, tuple)
    assert all(isinstance(x, int) for x in result)



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Check if a list of coordinates follows (lon, lat) or (lat, lon) """


def detect_coord_order(coords: list[tuple[float, float]]) -> str:
    """Checks whether a list of coordinate pairs follows (lon, lat) or (lat, lon).

    Args:
        coords (list[tuple[float, float]]): List of coordinate pairs to check.

    Returns:
        str: Returns "lon, lat", "lat, lon", or "ambiguous".
    """
    def is_lon_lat(x, y):
        return -180 <= x <= 180 and -90 <= y <= 90

    def is_lat_lon(x, y):
        return -90 <= x <= 90 and -180 <= y <= 180

    lonlat_count = 0
    latlon_count = 0
    ambiguous_count = 0

    for x, y in coords:
        # Check each ordering exclusively
        lonlat_valid = is_lon_lat(x, y) and not is_lat_lon(x, y)
        latlon_valid = is_lat_lon(x, y) and not is_lon_lat(x, y)

        if lonlat_valid:
            lonlat_count += 1
        elif latlon_valid:
            latlon_count += 1
        else:
            ambiguous_count += 1

    # Decide based on majority or ambiguity
    if lonlat_count > latlon_count and lonlat_count >= ambiguous_count:
        return "lon, lat"
    elif latlon_count > lonlat_count and latlon_count >= ambiguous_count:
        return "lat, lon"

    return "ambiguous"



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Utilities for autonomous vehicle simulation in SUMO."""

from __future__ import absolute_import
import colorsys
import xml.etree.ElementTree as ET
import traci
import os
import sys
from pathlib import Path
from xml.dom import minidom

if 'SUMO_HOME' in os.environ:
    sys.path.append(os.path.join(os.environ['SUMO_HOME'], 'tools'))
sys.path.append(os.path.join('c:', os.sep, 'whatever',
                'path', 'to', 'sumo', 'tools'))


def prettify_xml(xml_tree: ET.ElementTree) -> str:
    """Return a pretty-printed XML string for the ElementTree."""

    xml_tree_str = ET.tostring(xml_tree.getroot(), encoding='utf-8', method='xml')
    parsed_xml = minidom.parseString(xml_tree_str)
    return parsed_xml.toprettyxml(indent="    ")


def generate_rgb_colors(num_colors: int) -> list:
    """Generate `n` visually distinct RGB colors.
    Returns a list of (r, g, b) tuples with values in 0-255.
    """
    colors = []
    for i in range(num_colors):
        # evenly spaced hue around [0, 1)
        h = i / num_colors
        s = 1.0      # full saturation
        v = 1.0      # full brightness
        # convert to floats in [0,1]
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        # rescale to 0–255 ints
        colors.append(f"{int(r * 255)},{int(g * 255)},{int(b * 255)}")
    return colors


# Function to generate dictionary with all vehicle type attributes needed to create new vehicle type in SUMO
def create_veh_type_attributes(av_config: dict) -> dict:
    """ Create a dictionary with all vehicle type attributes needed to create new vehicle types in SUMO."""

    vTypeAttributes = ["id", "vClass",
                       "carFollowModel", "laneChangeModel",
                       "probability",
                       "accel", "decel", "minGap", "sigma", "tau", "emergencyDecel"]

    # Extract values from the configuration
    veh_types = av_config.get("veh_types", [])
    CFmodel_USER = av_config.get("CFmodel", {})
    pct_penetration = av_config.get("pct_penetration")
    car_follow_model = av_config.get("car_follow_model", "Krauss")
    lane_change_model = av_config.get("lane_change_model", "LC2013")

    # Create a dictionary for vehicle type attributes
    veh_type_attributes = {vType: dict.fromkeys(vTypeAttributes) for vType in veh_types}
    for i, vType in enumerate(veh_types):
        veh_type_attributes[vType]['id'] = vType
        veh_type_attributes[vType]['vClass'] = "passenger"
        veh_type_attributes[vType]['carFollowModel'] = car_follow_model
        veh_type_attributes[vType]['laneChangeModel'] = lane_change_model
        veh_type_attributes[vType]['probability'] = pct_penetration[i] * 0.01

        for i in range(5, len(vTypeAttributes)):
            veh_type_attributes[vType][vTypeAttributes[i]] = CFmodel_USER[vType][car_follow_model][vTypeAttributes[i]]

    return veh_type_attributes


# # Incorporate new vehicle types and vehicle type distributions in SUMO flow file to create a new modified flow file
def update_sumo_flow_xml(path_flow: str, veh_types: list, veh_type_attributes: dict):
    """ Update the SUMO flow XML file with new vehicle types and their attributes."""

    # Create the root element
    root = ET.Element("routes")

    # Create child elements
    vTypeDistribution = ET.SubElement(root, "vTypeDistribution")
    vTypeDistribution.set("id", "vdis")

    num_veh_types = len(veh_types)
    colors_list = generate_rgb_colors(num_veh_types)
    for i, vType in enumerate(veh_types):
        keysList = list(veh_type_attributes[vType].keys())
        vType_new = ET.SubElement(vTypeDistribution, "vType")
        for key in keysList:
            vType_new.set(key, str(veh_type_attributes[vType][key]))
        vType_new.set("color", colors_list[i])

    # Create an XML file
    vTypeTree = ET.ElementTree(root)
    # flow_dir = Path(path_flow).parent
    # vTypeTree.write(flow_dir / "vTypeDistTree.xml")

    # Parse the existing flow XML file
    origFlowTree = ET.parse(path_flow)

    # Get all the elements from the previous flow XML file
    origFlowTreeElements = origFlowTree.findall('flow')

    # Iterate over the elements and add them to the second XML file
    for element in origFlowTreeElements:
        element.set('type', 'vdis')
        vTypeTree.getroot().append(element)

    # Write the second XML file to a new file
    vTypeTree.write(path_flow)
    print(f"  :flow file updated with new vehicle types and saved at: {path_flow}")
    return True


def create_sumo_rou_xml(path_net: str, path_flow: str, path_turn: str, path_rou: str) -> bool:
    """ Create a SUMO route file using the provided network, flow, and turn ratio files."""
    # os.system('jtrrouter --route-files='+filepath+'vTypeModifiedFlow.xml --turn-ratio-files='+filepath+'chattnew.turn.xml --net-file='+filepath+'chatt3.net.xml --output-file='+filepath+'chatt3_modified_test.rou.xml --accept-all-destinations 1 --vtype-output='+filepath+'vtype_output_test.xml')

    path_vtype_output = Path(path_rou).parent / 'vtype_output.xml'

    os.system(
        f'jtrrouter --route-files={path_flow} --turn-ratio-files={path_turn} --net-file={path_net} --output-file={path_rou} --accept-all-destinations 1 --vtype-output={str(path_vtype_output)}')
    print(f"  :Route File Created: {path_rou}")
    return True


def add_veh_types_to_rou(path_rou: str, veh_types: list, veh_type_attributes: dict) -> bool:
    """ Add vehicle types to the SUMO route file."""

    Tree = ET.parse(path_rou)
    root = Tree.getroot()

    num_veh_types = len(veh_types)
    colors_list = generate_rgb_colors(num_veh_types)
    for i, vType in enumerate(veh_types):
        keysList = list(veh_type_attributes[vType].keys())
        vType_new = ET.SubElement(root, "vType")
        for key in keysList:
            vType_new.set(key, str(veh_type_attributes[vType][key]))
        vType_new.set("color", colors_list[i])
        root.insert(0, vType_new)
    Tree.write(path_rou)
    color_str = ', '.join([f'{vType}: {colors_list[i]}' for i, vType in enumerate(veh_types)])
    print(f"  :Added Vehicle Types into Route File, color assigned: {color_str}")
    return True


def create_sumo_config(path_cfg: str,
                       sim_name: str = "chatt",
                       sim_start: int = 0,
                       sim_end: int = 3600):
    """ Create a SUMO configuration file with default settings.

    Args:
        path_cfg (str): The path where the configuration file will be saved.
        sim_name (str): The name of the simulation. Defaults to "chatt".
        sim_start (int): The start time of the simulation in seconds. Defaults to 0.
        sim_end (int): The end time of the simulation in seconds. Defaults to 3600.
    """

    # create the root element
    root = ET.Element("configuration")

    # input element
    input_elem = ET.SubElement(root, "input")

    input_net_file = ET.SubElement(input_elem, "net-file")
    input_net_file.set("value", f"{sim_name}.net.xml")

    input_rou_file = ET.SubElement(input_elem, "route-files")
    input_rou_file.set("value", f"{sim_name}.rou.xml")

    input_additional_files = ET.SubElement(input_elem, "additional-files")
    input_additional_files.set("value", "detector.add.xml")

    # output element
    output_elem = ET.SubElement(root, "output")

    output_full_output = ET.SubElement(output_elem, "full-output")
    output_full_output.set("value", f"{sim_name}_Full_Output.xml")

    output_amitran_output = ET.SubElement(output_elem, "amitran-output")
    output_amitran_output.set("value", f"{sim_name}_Amitran_Output.xml")

    output_vehroute_output = ET.SubElement(output_elem, "vehroute-output")
    output_vehroute_output.set("value", f"{sim_name}_VehRoute_Output.xml")

    # time element
    time_elem = ET.SubElement(root, "time")

    time_begin = ET.SubElement(time_elem, "begin")
    time_begin.set("value", f"{sim_start}")

    time_end = ET.SubElement(time_elem, "end")
    time_end.set("value", f"{sim_end}")  # 1 hour simulation

    time_step_length = ET.SubElement(time_elem, "step-length")
    time_step_length.set("value", "0.1")  # 100 ms time step

    # gui element
    gui_elem = ET.SubElement(root, "gui_only")
    gui_start = ET.SubElement(gui_elem, "start")
    gui_start.set("value", "t")  # start the GUI

    # report element
    report_elem = ET.SubElement(root, "report")

    report_no_warnings = ET.SubElement(report_elem, "no-warnings")
    report_no_warnings.set("value", "true")

    report_no_step_log = ET.SubElement(report_elem, "no-step-log")
    report_no_step_log.set("value", "true")

    # write the XML to the file
    tree = ET.ElementTree(root)

    formatted_string = prettify_xml(tree)
    with open(path_cfg, 'w', encoding='utf-8') as file:
        file.write(formatted_string)

    # tree.write(path_cfg, encoding='utf-8', xml_declaration=True)
    print(f"  :SUMO configuration file created at: {path_cfg}")
    return True


def run_sumo_simulation(path_cfg: str, sim_time: int = 3600) -> bool:
    """ Run a SUMO simulation using the provided configuration file.

    Args:
        path_cfg (str): the path to the SUMO configuration file.
        sim_time (int, optional): the duration of the simulation in seconds. Defaults to 3600.

    Returns:
        bool: True if the simulation ran successfully, False otherwise.
    """
    traci.start(["sumo-gui", "-c", f"{path_cfg}"])
    while traci.simulation.getTime() < sim_time:
        traci.simulationStep()
    traci.close()
    print(f"  :Simulation completed successfully. Output files are saved in: {Path(path_cfg).parent}")
    return True



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
"""Control of module imports for the RealTwin function library."""

import pyufunc as pf
import subprocess
import re

from realtwin.util_lib.find_exe_from_PATH import find_executable_from_PATH_on_win


# Check required simulation environments
def is_sumo_installed(*, ext: str = "exe", sel_dir: list = None, verbose: bool = True) -> bool | list:
    """Check if SUMO is installed on the system.

    Args:
        ext (str): The extension of the executable. Defaults to "exe" for executable files.
        sel_dir (list): A list of directories to search for the SUMO executable. Defaults to None.
        verbose (bool): Whether to print the process info. Defaults to True.

    Example:
        >>> import realtwin as rt
        >>> rt.is_sumo_installed(ext="exe", sel_dir=["C:/Program Files/SUMO/bin"])
        >>> True

    Raises:
        Exception: Unsupported OS, could not find SUMO executable

    Returns:
        bool or list: a list of installed versions, False otherwise.
    """

    # check the operation system
    if pf.is_windows():
        print("  :Checking SUMO installation on Windows.")
        sumo_executable = "sumo.exe"  # For Windows

    elif pf.is_linux():
        print("  :Checking SUMO installation on Linux.")
        sumo_executable = None  # TODO: Check name of the executable

    elif pf.is_mac():
        print("  :Checking SUMO installation on MacOS.")
        sumo_executable = None  # TODO: Check name of the executable

    else:
        raise Exception("  :Unsupported OS, could not find SUMO executable.")

    # Check if 'sumo' executable is in PATH: return None if not found
    sumo_exe_lst = find_executable_from_PATH_on_win(sumo_executable, ext=ext, sel_dir=sel_dir, verbose=False)

    if sumo_exe_lst:
        # remove duplicates
        sumo_exe_lst = list(set(sumo_exe_lst))

        # print out the version of SUMO if more than one path is found
        if len(sumo_exe_lst) > 1:
            print("  :Multiple SUMO executables found in the system PATH:")

        # run SUMO to check the version
        version_lst = []
        for exe_path in sumo_exe_lst:
            try:
                version_check = subprocess.run([exe_path],
                                               capture_output=True,
                                               text=True,
                                               check=True)
                # find the version number from the output
                if version_check.returncode == 0:

                    # Define the pattern to match the version number
                    pattern = r'(\d+\.\d+\.\d+)'

                    # Search for the pattern in the text
                    match = re.search(pattern, version_check.stdout)

                    # Extract and print the version number if a match is found
                    if match:
                        version = match.group(1)
                        version_lst.append(str(version))
                        print(f"  :SUMO version: {version} found: {exe_path}")
                    else:
                        pass

            except Exception as e:
                if verbose:
                    print(f"  :Error running SUMO: {e}")
        if version_lst:
            return list(set(version_lst))

    print("  :SUMO not found in the system PATH.")
    return False


def is_vissim_installed(*, ext: str = "", sel_dir: list = None, verbose: bool = True) -> bool:
    """Check if VISSIM is installed on the system.

    Args:
        ext (str): The extension of the executable. Defaults to "".
        sel_dir (list): A list of directories to search for the VISSIM executable. Defaults to None.
        verbose (bool): Whether to print the process info. Defaults to True.

    Returns:
        bool: True if VISSIM is installed, False otherwise.
    """
    print("  Warning: Checking VISSIM installation is not supported yet.")
    return False


def is_aimsun_installed(*, ext: str = "", sel_dir: list = None, verbose: bool = True) -> bool:
    """Check if AIMSUN is installed on the system.

    Args:
        ext (str): The extension of the executable. Defaults to "".
        sel_dir (list): A list of directories to search for the AIMSUN executable. Defaults to None.
        verbose (bool): Whether to print the process info. Defaults to True.

    Returns:
        bool: True if AIMSUN is installed, False otherwise.
    """
    print("  Warning: Checking AIMSUN installation is not supported yet.")
    return False



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Check if the code is running in a Jupyter Notebook environment """

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from IPython import get_ipython


def is_running_in_notebook():
    """ Check if the code is running in a Jupyter Notebook environment """
    try:
        shell = get_ipython().__class__.__name__
        return shell == 'ZMQInteractiveShell'
        # if shell == 'ZMQInteractiveShell':
        #     return True  # Running in a Jupyter Notebook or IPython
        # return False  # Running in a different IPython environment (e.g., IPython shell)
    except NameError:
        return False



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host Network element of Abstract scenario
'''
from pathlib import Path
import osmnx as ox
import networkx as nx
import pandas as pd
import numpy as np
import os
import shutil
from shapely.geometry import Polygon
import math
import pyufunc as pf


class Network:
    """The Network class to host the Network element of Abstract scenario
    """
    def __init__(self, output_dir: str = None):
        self._output_dir = output_dir
        self.NetworkName = "network"
        self.NetworkVertices = ""
        self.ElevationMap = "No elevation map provided!"
        self.OpenDriveNetwork = OpenDriveNetwork(output_dir=self._output_dir)

    def isEmpty(self):
        """Check if the Network element is empty"""
        pass


class OpenDriveNetwork:
    """The OpenDriveNetwork class to generate OpenDrive network"""
    def __init__(self,
                 net_name: str = "network",
                 net_vertices: str = "",
                 ele_map: str = "No elevation map provided!",
                 output_dir: str = "output"):

        self._net_name = net_name
        self._net_vertices = net_vertices
        self._ele_map = ele_map

        self._output_dir = output_dir
        self.OpenDrive_network = []

        # create output directory
        path_openDrive = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive'))

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir, exist_ok=True)

        # Create OpenDrive directory
        if os.path.exists(path_openDrive):
            # Delete existing directory and its contents
            shutil.rmtree(path_openDrive)
        # Create new directory
        os.makedirs(path_openDrive, exist_ok=True)

    def isEmpty(self):
        """Check if the OpenDriveNetwork element is empty"""
        pass

    def setValue(self):
        """Create OpenDrive network and save it to the output directory"""

        path_openDrive = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive'))

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir, exist_ok=True)

        if os.path.exists(path_openDrive):
            # Delete existing directory and its contents
            shutil.rmtree(path_openDrive)

        # Create new directory
        os.makedirs(path_openDrive, exist_ok=True)

        self.create_SUMO_network()
        self.create_OpenDrive_network()

        # self.OpenDrive_network = pf.path2linux(f'{path_openDrive}/{self._net_name}.xodr')

        # delete intermediate files         ################################################################################# add back
        # PathDelete1 = os.path.join(OpenDrivePath, '{}.net.xml'.format(Name))
        # PathDelete2 = os.path.join(OpenDrivePath, '{}_WithElevation.net.xml'.format(Name))
        # PathDelete3 = os.path.join(OpenDrivePath, 'edges.edg.xml')
        # PathDelete4 = os.path.join(OpenDrivePath, 'nodes.nod.xml')
        # if os.path.exists(PathDelete1):
        #     os.remove(PathDelete1)  # Delete the file
        # if os.path.exists(PathDelete2):
        #     os.remove(PathDelete2)  # Delete the file
        # if os.path.exists(PathDelete3):
        #     os.remove(PathDelete3)  # Delete the file
        # if os.path.exists(PathDelete4):
        #     os.remove(PathDelete4)  # Delete the file

    def create_OpenDrive_network(self):
        """Create OpenDrive network from SUMO Network"""
        # get node and edge path
        path_net = pf.path2linux(os.path.join(self._output_dir, f'OpenDrive/{self._net_name}.net.xml'))
        path_net_elevation = pf.path2linux(
            os.path.join(self._output_dir, f'OpenDrive/{self._net_name}_WithElevation.net.xml'))
        path_open_drive_ele = pf.path2linux(
            os.path.join(self._output_dir, f'OpenDrive/{self._net_name}_WithElevation.xodr'))
        path_open_drive_output = pf.path2linux(
            os.path.join(self._output_dir, f'OpenDrive/{self._net_name}.xodr'))

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir, exist_ok=True)

        if not os.path.exists(Path(self._output_dir) / 'OpenDrive'):
            os.makedirs(Path(self._output_dir) / 'OpenDrive', exist_ok=True)

        command2 = f'cmd/c "netconvert -s {path_net} --opendrive-output={path_open_drive_output}"'
        os.system(command2)
        self.OpenDrive_network = [path_open_drive_output]

        # with elevation data
        if self._ele_map != "No elevation map provided!":
            command4 = f'cmd/c "netconvert -s {path_net_elevation} --opendrive-output={path_open_drive_ele}"'
            os.system(command4)
            self.OpenDrive_network.append(path_open_drive_ele)
        else:
            self.OpenDrive_network.append(None)

    def create_SUMO_network(self):
        """Generate OpenDrive network from OpenStreetMap"""

        path_openDrive = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive'))

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir, exist_ok=True)

        if os.path.exists(path_openDrive):
            # Delete existing directory and its contents
            shutil.rmtree(path_openDrive)

        # Create new directory
        os.makedirs(path_openDrive, exist_ok=True)

        # create networkx graph network
        nx_net = OSMRoad(output_dir=self._output_dir)
        nx_net.get_graph(self._net_vertices)
        nx_net.add_spread()
        nx_net.add_num_lane()
        nx_net.add_missing_value()
        nx_net.generate_edges()
        nx_net.generate_nodes()

        # get node and edge path
        path_node = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive/nodes.nod.xml'))
        path_edge = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive/edges.edg.xml'))
        path_net = pf.path2linux(os.path.join(self._output_dir, f'OpenDrive/{self._net_name}.net.xml'))
        path_net_elevation = pf.path2linux(
            os.path.join(self._output_dir, f'OpenDrive/{self._net_name}_WithElevation.net.xml'))

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir, exist_ok=True)

        if not os.path.exists(Path(self._output_dir) / 'OpenDrive'):
            os.makedirs(Path(self._output_dir) / 'OpenDrive', exist_ok=True)

        # without elevation data
        command1 = f'cmd/c "netconvert --node-files={path_node} --edge-files={path_edge}\
        --output-file={path_net} --roundabouts.guess --ramps.guess\
        --tls.discard-simple --tls.join --proj.utm --ignore-errors.edge-type"'

        # command2 = f'cmd/c "netconvert -s {path_net} --opendrive-output={path_open_drive_output}"'

        os.system(command1)
        # os.system(command2)
        # self.OpenDrive_network = [path_open_drive_output]

        # with elevation data
        if self._ele_map != "No elevation map provided!":
            command3 = f'cmd/c "netconvert --node-files={path_node} --edge-files={path_edge}\
            --output-file={path_net_elevation} --roundabouts.guess --ramps.guess\
            --tls.discard-simple --tls.join --heightmap.geotiff {self._net_name}.tif --proj.utm"'

            # command4 = f'cmd/c "netconvert -s {path_net_elevation} --opendrive-output={path_open_drive_ele}"'

            os.system(command3)
            # os.system(command4)
            # self.OpenDrive_network.append(path_open_drive_ele)
        else:
            # self.OpenDrive_network.append(None)
            pass


class OSMRoad:
    """The OSMRoad class to generate network from OpenStreetMap"""

    def __init__(self, output_dir : str = "RT_Network"):

        self._output_dir = output_dir
        self.G = nx.empty_graph()

    def get_graph(self, NetworkVertices):
        """Get the graph from OpenStreetMap using osmnx"""

        if isinstance(NetworkVertices, str):

            # NetworkVertices = " (-85.14985634615098, 35.0401442512819), ( -85.15827848213861, 35.04329333848216),
            # (-85.15818728703556,35.04347780007786), (-85.14975978663011,35.04037264104675)"

            NetworkVerticesTemp = NetworkVertices.replace(' ', '').split('),(')
            NetworkVerticesTemp[0] = NetworkVerticesTemp[0][1:]
            NetworkVerticesTemp[-1] = NetworkVerticesTemp[-1][:-1]
            NetworkVerticesList = [tuple((float(ele) for ele in tup.split(',')))
                                   for tup in NetworkVerticesTemp]
        elif isinstance(NetworkVertices, list):
            # Check if the list contains tuples
            if all(isinstance(item, (list, tuple)) and len(item) == 2 for item in NetworkVertices):
                NetworkVerticesList = [(float(item[0]), float(item[1])) for item in NetworkVertices]
            else:
                raise ValueError("Invalid format: List must contain list of [lon, lat].")

        polygon = Polygon(NetworkVerticesList)
        self.G = ox.graph.graph_from_polygon(
            polygon, network_type='drive_service', simplify=True, truncate_by_edge=True)
        self.G = ox.add_edge_speeds(self.G)
        self.G = ox.add_edge_travel_times(self.G)

        self.G.remove_edges_from(nx.selfloop_edges(self.G))

        nodes, edges = ox.graph_to_gdfs(self.G, fill_edge_geometry=True)

        self.G = ox.graph_from_gdfs(nodes, edges, graph_attrs=self.G.graph)
        # ox.plot_graph(self.G)

    def line_str(self, line):
        """Convert line to string"""

        for jx, j in enumerate(list(line.coords)):
            # s_x,s_y=net.convertLonLat2XY(j[0],j[1])
            if jx == 0:
                # s = '%s,%s' % (j[0], j[1])
                s = f"{j[0]},{j[1]}"
            else:
                # s = s + ' %s,%s' % (j[0], j[1])
                s += f" {j[0]},{j[1]}"
        return s

    def split_lane_no(self, str_digit):
        """Split the lane number"""
        float_digit = float(str_digit)
        return [math.ceil(float_digit / 2.0), math.floor(float_digit / 2.0)]

    def add_spread(self):
        """Add spread to the edges"""

        for _, _, d in self.G.edges(data=True):
            if 'lanes' in d:
                try:
                    # no_lanes = '%.0f' % (float(d['lanes']))
                    no_lanes = f"{float(d['lanes']):.0f}"
                except Exception:
                    # no_lanes = '%.0f' % (float(d['lanes'][0]))
                    no_lanes = f"{float(d['lanes'][0]):.0f}"

                if d['oneway']:
                    d['spread'] = 'center'
                    d['n_lane'] = no_lanes
                else:
                    if float(no_lanes) > 1 and float(no_lanes) % 2 == 0:
                        d['spread'] = 'right'
                        d['n_lane'] = self.split_lane_no(no_lanes)

                    elif float(no_lanes) > 1 and float(no_lanes) % 2 == 0:
                        d['spread'] = 'roadCenter'
                        d['n_lane'] = self.split_lane_no(no_lanes)

    def add_num_lane(self):
        """Add number of lanes to the edges"""
        for e1, e2, d in self.G.edges(data=True):
            if 'n_lane' in d:
                if d['oneway'] is False and isinstance(d['n_lane'], list) is True:
                    k = list(self.G[e2][e1].items())[0][0]
                    self.G[e2][e1][k]['n_lane'] = f'{float(d["n_lane"][-1]):.0f}'

                    # d['n_lane'] = '%.0f' % (float(d['n_lane'][0]))
                    d["n_lane"] = f"{float(d['n_lane'][0]):.0f}"

    def add_missing_value(self):
        """Add missing values to the edges"""
        for e1, e2, d in self.G.edges(data=True):
            # d['u_id'] = '%s%s' % (e1, e2)
            d["u_id"] = f"{e1}{e2}"

            if 'n_lane' not in d:
                d['n_lane'] = ' '

            if 'spread' not in d:
                if d['oneway']:
                    d['spread'] = 'center'
                else:
                    d['spread'] = 'right'

    def generate_edges(self):
        """Generate edges from the graph"""
        edges = []
        for e1, e2, d in self.G.edges(data=True):
            edges.append([d['u_id'], e1, e2, f'{d["n_lane"]}',
                          self.line_str(d['geometry']), float(
                              d['speed_kph']) * 0.28 + 2.22,
                          d['spread']])

        df_edge = pd.DataFrame(edges)

        df_edge.columns = ['id', 'from', 'to', 'numLanes', 'shape', 'speed', 'spreadType']
        for i in df_edge.index:
            if df_edge['numLanes'][i] == ' ':
                df_edge.at[i, 'numLanes'] = np.nan

        # df_edge=df_edge.fillna(np.nan)
        # print(df_edge.head())

        path_edge = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive/edges.edg.xml'))
        df_edge.to_xml(path_edge,
                       attr_cols=['id', 'from', 'to', 'numLanes',
                                  'shape', 'speed', 'spreadType'],
                       root_name='edges',
                       row_name='edge',
                       index=False)

    def generate_nodes(self):
        """Generate nodes from the graph"""

        nodes = []
        for n, d in self.G.nodes(data=True):
            nodes.append([n, d['x'], d['y'], 'priority'])
            try:
                if d['highway'] == 'traffic_signals':
                    nodes.append([n, d['x'], d['y'], 'traffic_light'])
            except Exception:
                continue

        df_node = pd.DataFrame(nodes)
        df_node.columns = ['id', 'x', 'y', 'type']

        path_node = pf.path2linux(os.path.join(self._output_dir, 'OpenDrive/nodes.nod.xml'))
        df_node.to_xml(path_node,
                       attr_cols=['id', 'x', 'y', 'type'],
                       root_name='nodes',
                       row_name='node',
                       index=False)



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Generate loop detector .add.xml file for SUMO."""

from pathlib import Path
from datetime import datetime
from xml.dom import minidom
import xml.etree.ElementTree as ET


def xml_prettify(element: str) -> str:
    """Return a pretty-printed XML string for the Element."""

    rough_string = ET.tostring(element, 'utf-8')
    re_parsed = minidom.parseString(rough_string)
    return re_parsed.toprettyxml(indent="    ")


def generate_sumo_loop_detector_add_xml(path_net: str | Path, *,
                                        detector_type: str = "E1",
                                        add_fname: str = "detector.add.xml",
                                        detector_output_fname: str = "",
                                        dest_dir: str = "") -> bool:
    """Generate the .add.xml file for SUMO and add loop detectors for each lane that has a detector.

    Args:
        path_net (str | Path): Path to the SUMO network file (.net.xml).
        detector_type (str): The type of detector to be added. Defaults to "E1".
            Accepted types: E1: Inductive loop detector, E2: Lane area detector, E0: Instant induction loops.
        add_fname (str): SUMO additional file. Defaults to "detector.add.xml".
        detector_output_fname (str): The output file name to record loop detectors data in simulation.
            the output file name in default is detector_output_YYYYMMDDHHMM.xml
        dest_dir (str): The destination directory to save the .add.xml file. Defaults to the current directory.

    See Also:
        For different detector types, please refer to the SUMO documentation:
            https://sumo.dlr.de/docs/Simulation/Output/#simulated_detectors

        E1: Inductive loop detector
            https://sumo.dlr.de/docs/Simulation/Output/Induction_Loops_Detectors_%28E1%29.html

        E2: Lane area detector,
            https://sumo.dlr.de/docs/Simulation/Output/Lanearea_Detectors_%28E2%29.html

        E0: Instant induction loops,
            https://sumo.dlr.de/docs/Simulation/Output/Instantaneous_Induction_Loops_Detectors.html
    """

    # get all lanes from the SUMO network
    net = ET.parse(path_net)
    root = net.getroot()

    # get all traffic signal id
    tlLights = root.findall(".//tlLogic")
    tl_ids = [tl.get("id") for tl in tlLights]

    # assume tl_ids are signalized intersection ids

    edges = root.findall(".//edge")
    # edges_not_internal = [edge for edge in edges if edge.get("function") != "internal"]
    edges_not_internal = edges

    lane_lookup_dict = {}
    for edge in edges_not_internal:
        # if the to id of the edge is in tl_ids, add the lane to the lookup dict
        if edge.get("to") in tl_ids:
            for lane in edge.findall(".//lane"):
                lane_id = lane.get("id")
                lane_length = float(lane.get("length"))
                lane_lookup_dict[lane_id] = {
                    "length": lane_length,
                    "numDetects": 1
                }

    # get detector tag
    if detector_type == "E1":
        detector_tag = "inductionLoop"
    elif detector_type == "E2":
        detector_tag = "laneAreaDetector"
    elif detector_type == "E0":
        detector_tag = "instantInductionLoop"
    else:
        raise ValueError(
            f"Unknown detector type: {detector_type}. Accepted types are E1, E2, E0.")

    add_elem = ET.Element("additional")

    if not str(add_fname).endswith(".add.xml"):
        add_fname = f"{add_fname}.add.xml"

    if detector_output_fname:
        if not str(detector_output_fname).endswith(".xml"):
            detector_output_fname = f"{detector_output_fname}.xml"
    else:
        detector_output_fname = f"detector_output_{datetime.now().strftime(r'%Y%m%d')}.xml"

    for lane_id, lane_info in lane_lookup_dict.items():
        # Check if the lane has a detector, num_detectors not None
        if lane_info.get("numDetects"):
            # Create the detector element
            detector = ET.SubElement(add_elem, detector_tag)
            detector.set("id", f"{lane_id}_detector")
            detector.set("lane", f"{lane_id}")
            detector.set("pos", "-8")  # must assigned, backward of the lane
            detector.set("friendlyPos", "true")
            # detector.set("vTypes", "")
            detector.set("file", f"{detector_output_fname}")  # output file name

            # # Add the detector to the root element and save root file to path_net
            # root = ET.Element("additional")
            # root.append(detector)

    xml_str = xml_prettify(add_elem)
    path_output = Path(dest_dir) / add_fname if dest_dir else Path(add_fname)
    with open(path_output, "w", encoding="utf-8") as f:
        f.write(xml_str)
    return True


# if __name__ == "__main__":
#     pass

    # path_net_ = Path(r"../../datasets/avtes/chatt3_updated_signal_v2.net.xml")
    # detector_type_ = "E1"
    # add_fname_ = "../datasets/avtes/detector.add.xml"
    # sim_output_fname_ = "detector_output.xml"
    # generate_sumo_loop_detector_add_xml(path_net_, detector_type_, add_fname_, sim_output_fname_)



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""control of module imports for the RealTwin package."""

from .create_venv import venv_create, venv_delete
from .download_elevation_tif import download_elevation_tif_by_bbox
from .download_file_from_web import download_single_file_from_web
from .find_exe_from_PATH import find_executable_from_PATH_on_win
from .create_config import prepare_realtwin_configs
from .get_bbox_from_list_of_coords import get_bbox_from_vertices

__all__ = [

    # create_venv
    'venv_create',
    'venv_delete',

    'download_elevation_tif_by_bbox',
    'download_single_file_from_web',
    'find_executable_from_PATH_on_win',

    'prepare_realtwin_configs',
    'get_bbox_from_vertices',
]



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Find the executable from the system PATH """

import os
import subprocess
import pyufunc as pf


def find_executable_from_PATH_on_win(exe_name: str,
                                     ext: str = "exe",
                                     sel_dir: list = None,
                                     verbose: bool = True) -> list | None:
    """Find the executable from the system PATH.

    Args:
        exe_name (str): The executable name to search for.
        ext (str): The extension of the executable. Defaults to "exe" for executable files.
        sel_dir (list): A list of directories to search for the executable. Defaults to [].
        verbose (bool): Whether to print the process info. Defaults to True.

    Returns:
        list or None: A list of full path of the executable if found, otherwise None.
    """

    # check if the executable name is a string
    if not isinstance(exe_name, str):
        raise ValueError("exe_name should be a string.")

    # check if extension is str
    if not isinstance(ext, str):
        raise ValueError("ext should be a string.")

    # check if sel_dir is a list
    if not isinstance(sel_dir, (list, type(None))):
        raise ValueError("sel_dir should be a list.")

    # Add the selected directories to the PATH environment
    if sel_dir:
        for path in sel_dir:
            if os.path.isdir(path):
                os.environ["PATH"] += os.pathsep + path
            elif verbose:
                print(f"  :The directory: {path} does not exist. Skipped.")

    # check if exe_name has the extension
    _, ext_str = os.path.splitext(exe_name)
    if not ext_str:
        if verbose:
            print(f"  :The executable: {exe_name} has no extension. Added {ext} as the extension.")
        exe_name = f"{exe_name}.{ext}"

    # get the full environment PATH
    env_paths = os.getenv("PATH").split(os.pathsep)

    res = []
    for path in env_paths:
        abs_path = pf.path2linux(os.path.join(path, exe_name))

        # check if the file exists and is executable
        if os.path.isfile(abs_path) and os.access(abs_path, os.X_OK):
            res.append(abs_path)

    if not res:
        if verbose:
            print(f"  :Could not find {exe_name} in the system PATH."
                  " Please make sure the executable is installed."
                  " please include executable extension in the name.")
        return None

    if verbose:
        print(f"  :Found {exe_name} in the system PATH:")
        for path in res:
            print(f"    :{path}")
    return res


def find_executable_from_PATH_on_linux(exe_name: str,
                                       verbose: bool = True) -> list[str]:
    """Use the system `which -a` to list all matches for exe_name on Linux.

    Args:
        exe_name (str): The name of the executable to search for.
        verbose (bool): Whether to print the process info. Defaults to True.

    Example:
        >>> find_executable_from_PATH_on_linux("python3", verbose=True)
        >>> ['/usr/bin/python3', '/usr/local/bin/python3']

    Returns:
        Optional[List[str]]: A list of paths where the executable is found, or None if not found.
    """

    try:
        # -a: list all matches, not just the first
        proc = subprocess.run(
            ["which", "-a", exe_name],
            capture_output=True,
            text=True,
            check=False)
        paths = [line.strip() for line in proc.stdout.splitlines() if line.strip()]
        if verbose:
            if paths:
                for p in paths:
                    print(f"[which] Found: {p}")
            else:
                print(f"[which] No matches for {exe_name}")
        return paths or None

    except FileNotFoundError:
        if verbose:
            print("[which] `which` command not found on this system.")
        return None



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Behavior Calibration class for SUMO """

import os
import sys
from functools import partial
import random
from pathlib import Path
import subprocess
import warnings
import pyufunc as pf

from mealpy import FloatVar, SA, GA, TS
try:
    from realtwin.func_lib._f_calibration.algo_sumo.util_cali_behavior import (
        get_travel_time_from_EdgeData_xml,
        update_flow_xml_from_solution,
        run_jtrrouter_to_create_rou_xml,
        # result_analysis_on_EdgeData,
    )
except ImportError:
    from util_cali_behavior import (
        get_travel_time_from_EdgeData_xml,
        update_flow_xml_from_solution,
        run_jtrrouter_to_create_rou_xml,
        # result_analysis_on_EdgeData,
    )
import numpy as np

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
    sys.path = list(set(sys.path))  # remove duplicates

else:
    warnings.warn("Environment variable 'SUMO_HOME' is not set. "
                  "please declare environment variable 'SUMO_HOME'")
    # sys.exit("please declare environment variable 'SUMO_HOME'")

rng = np.random.default_rng(seed=812)


def fitness_func(solution: list | np.ndarray, scenario_config: dict = None, error_func: str = "rmse") -> float:
    """ Evaluate the fitness of a given solution for SUMO calibration."""

    # Set up SUMO command with car-following parameters
    if error_func not in ["rmse", "mae"]:
        raise ValueError("error_func must be either 'rmse' or 'mae'")

    if solution[5] >= 9.3:  # emergencyDecel
        solution[5] = 9.3
    if solution[5] < solution[2]:  # emergencyDecel < deceleration
        solution[5] = solution[2] + random.randrange(1, 5)

    # get path from scenario_config
    network_name = scenario_config.get("network_name")
    sel_behavior_routes = scenario_config.get("sel_behavior_routes")
    sim_input_dir = Path(scenario_config.get("dir_behavior"))
    path_net = pf.path2linux(sim_input_dir / f"{network_name}.net.xml")
    path_flow = pf.path2linux(sim_input_dir / f"{network_name}.flow.xml")
    path_turn = pf.path2linux(sim_input_dir / f"{network_name}.turn.xml")
    path_rou = pf.path2linux(sim_input_dir / f"{network_name}.rou.xml")
    path_EdgeData = pf.path2linux(sim_input_dir / "EdgeData.xml")

    sim_name = scenario_config.get("sim_name")
    # sim_end_time = scenario_config.get("sim_end_time")

    # change the working directory to the input directory for SUMO
    os.chdir(sim_input_dir)

    update_flow_xml_from_solution(path_flow, solution)

    # Create the route XML file using jtrrouter
    run_jtrrouter_to_create_rou_xml(network_name, path_net, path_flow, path_turn, path_rou)

    # Define the command to run SUMO
    sumo_command = f"sumo -c \"{sim_name}\""
    sumoProcess = subprocess.Popen(sumo_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    sumoProcess.wait()

    # Read output file or TraCI to evaluate the fitness
    travel_time_diff_list = []
    for route in sel_behavior_routes:
        route_time = sel_behavior_routes[route]["time"]
        route_edge_list = sel_behavior_routes[route]["route_list"]
        travel_time = get_travel_time_from_EdgeData_xml(path_EdgeData, route_edge_list)
        travel_time_diff = abs(route_time - travel_time)
        travel_time_diff_list.append(travel_time_diff)

    if error_func == "rmse":
        fitness_err = np.sqrt(sum([x ** 2 for x in travel_time_diff_list]) / len(travel_time_diff_list))
    elif error_func == "mae":
        fitness_err = sum(travel_time_diff_list) / len(travel_time_diff_list)
    else:
        raise ValueError("error_func must be either 'rmse' or 'mae'")

    # Calculate GEH from updated results
    # path_summary = pf.path2linux(sim_input_dir / "summary.xlsx")
    # calibration_target = scenario_config.get("calibration_target")
    # sim_start_time = scenario_config.get("sim_start_time")
    # sim_end_time = scenario_config.get("sim_end_time")
    # _, mean_geh, geh_percent = result_analysis_on_EdgeData(path_summary,
    #                                                        path_EdgeData,
    #                                                        calibration_target,
    #                                                        sim_start_time,
    #                                                        sim_end_time)
    # print(f"  :GEH: Mean Percentage: {mean_geh:.6f}, {geh_percent:.6f}, Travel time error: {fitness_err:.6f}")
    print(f"  :Travel time error: {fitness_err:.6f}")

    return fitness_err


class BehaviorCali:
    """ Behavior Optimization class for SUMO calibration

    Args:
        scenario_config (dict): the configuration for the scenario.
        behavior_config (dict): the configuration for the behavior.
        verbose (bool): whether to print the log. Defaults to True.

    Notes:
        We use the mealpy library for optimization. mealpy is a Python library for optimization algorithms.
            https://mealpy.readthedocs.io/en/latest/index.html

        1. The `init_solution` parameter is used to provide initial solutions for the population. None by default.
        2. Behavior includes: min_gap(meters), acceleration(m/s^2), deceleration(m/s^2), sigma, tau, and emergencyDecel.

    See Also:
        Problem_dict: https://mealpy.readthedocs.io/en/latest/pages/general/simple_guide.html
        termination_dict: https://mealpy.readthedocs.io/en/latest/pages/general/advance_guide.html#stopping-condition-termination

    Examples:
        >>> from realtwin import BehaviorCali
        >>> prob_dict = {"obj_func": partial(fitness_func, scenario_config=scenario_config, error_func="rmse"),
                        "bounds": FloatVar(lb=[1.0, 2.5, 4, 0.0, 0.25, 5.0], ub=[3.0, 3.0, 5.3, 1.0, 1.25, 9.3],),
                        "minmax": "max",  # maximize or minimize
                        "log_to": "console",
                        "save_population": True}
        >>> init_solution = [2.5, 2.6, 4.5, 0.5, 1.0, 9.0]
        >>> term_dict = {"max_epoch": 500, "max_fe": 10000, "max_time": 3600, "max_early_stop": 20}
        >>> opt = BehaviorCali(problem_dict=prob_dict, init_solution=init_solution, term_dict=term_dict)
        >>> g_best, model_opt = opt.run_GA(epoch=1000, pop_size=30, pc=0.95, pm=0.1, sel_model="BaseGA")

        Save result figures to output_dir
        >>> opt.run_vis(output_dir="output_dir", model=model_opt)
        >>> print(g_best.solution)
        >>> print(g_best.target.fitness)
    """

    def __init__(self, scenario_config: dict = None, behavior_config: dict = None, verbose: bool = True):

        self.scenario_config = scenario_config
        self.behavior_cfg = behavior_config
        self.verbose = verbose

        # prepare termination criteria from scenario config
        self.term_dict = {
            "max_epoch": self.scenario_config.get("max_epoch", 1000),
            "max_fe": self.scenario_config.get("max_fe", 10000),
            "max_time": self.scenario_config.get("max_time", None),
            "max_early_stop": self.scenario_config.get("max_early_stop", 80),
        }

        init_params = self.behavior_cfg.get("initial_params", None)
        if isinstance(init_params, dict):
            self.init_solution = list(init_params.values())
        elif isinstance(init_params, list):
            self.init_solution = init_params
        elif isinstance(init_params, np.ndarray):
            self.init_solution = init_params.tolist()
        else:
            self.init_solution = None

        # Default parameter ranges
        params_ranges_ = {"min_gap": [1.0, 3.0],
                          "acceleration": [2.5, 3.0],
                          "deceleration": [4.0, 5.3],
                          "sigma": [0.0, 1.0],
                          "tau": [0.25, 1.25],
                          "emergencyDecel": [5.0, 9.3]
                          }

        params_ranges = self.behavior_cfg.get("params_ranges",
                                              params_ranges_).values()
        params_lb = [val[0] for val in params_ranges]
        params_ub = [val[1] for val in params_ranges]
        self.problem_dict = {
            "obj_func": partial(fitness_func, scenario_config=self.scenario_config, error_func="rmse"),
            "bounds": FloatVar(lb=params_lb, ub=params_ub,),
            "minmax": "min",  # maximize or minimize
            "log_to": "console",
            # "log_to": "file",
            # "log_file": "result.log",
            "save_population": True,              # Default = False
        }

    def _generate_initial_solutions(self, init_vals: list, pop_size: int) -> np.array:
        """Generate initial solutions for inputs.

        Args:
            init_vals (list | np.array): initial values for the solutions.
            pop_size (int): population size.

        Returns:
            np.array: array of initial solutions.
        """

        # TDD
        if not isinstance(init_vals, (list, np.ndarray)):
            print("Error: init_vals must be a list, numpy array, or None.")
            return None

        if init_vals is not None:
            return np.array(list(init_vals) * pop_size).reshape(pop_size, len(init_vals))
        return None

    def run_vis(self, output_dir: str, model) -> bool:
        """ Save the results of the optimization.

        See Also:
            https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.utils.html#module-mealpy.utils.history

        Args:
            output_dir (str): the directory to save the results.
            model: the optimized model object.
        """

        # save the best solution
        try:
            pass
            # model.history.save_global_objectives_chart(filename=f"{output_dir}/global_objectives")
            # model.history.save_local_objectives_chart(filename=f"{output_dir}/local_objectives")
            # model.history.save_global_best_fitness_chart(filename=f"{output_dir}/global_best_fitness")
            # model.history.save_local_best_fitness_chart(filename=f"{output_dir}/local_best_fitness")
            # model.history.save_runtime_chart(filename=f"{output_dir}/runtime")
            # model.history.save_exploration_exploitation_chart(filename=f"{output_dir}/exploration_exploitation")
            # model.history.save_diversity_chart(filename=f"{output_dir}/diversity")
            # model.history.save_trajectory_chart(filename=f"{output_dir}/trajectory")
        except Exception as e:
            print(f"  :Error in saving vis: {e}")
            return False
        return True

    def run_GA(self, **kwargs):
        """Run Genetic Algorithm (GA) for behavior optimization.

        Note:
            1. The `model_selection` parameter allows you to choose different types of GA models. Default is "BaseGA".
                Options include "BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", and "SingleGA".
            2. Additional keyword arguments (`**kwargs`) can be passed for specific GA models (See Also).
            3. Please check original GA model documentation for more kwargs in details: https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.evolutionary_based.html#module-mealpy.evolutionary_based.GA

        Warning:
            You can change the input parameters only from input_config.yaml file.

        See Also:
            https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.evolutionary_based.html#module-mealpy.evolutionary_based.GA

        Notes:
            GA parameters:
                epoch (int): the iterations. Defaults to 1000.
                pop_size (int): population size. Defaults to 50.
                pc (float): crossover probability. Defaults to 0.95.
                pm (float): mutation probability. Defaults to 0.025.
                model_selection (str): the type of GA model to use. Defaults to "BaseGA".
                    options: "BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", "SingleGA".

        Args:
            kwargs: additional keyword arguments for specific GA models (Navigate to See Also).
        """
        if (ga_config := self.behavior_cfg.get("ga_config")) is None:
            raise ValueError("ga_config is not provided in Calibration/turn_inflow setting in input_config.yaml file.")

        epoch = ga_config.get("epoch", 1000)  # max iterations
        pop_size = ga_config.get("pop_size", 50)  # population size
        pop_size = max(pop_size, 10)  # ensure population size is at least 10
        pc = ga_config.get("pc", 0.75)  # crossover probability
        pm = ga_config.get("pm", 0.1)  # mutation probability

        selection = ga_config.get("selection", "roulette")  # selection method
        k_way = ga_config.get("k_way", 0.2)  # k-way for tournament selection
        crossover = ga_config.get("crossover", "uniform")  # crossover method
        mutation = ga_config.get("mutation", "swap")  # mutation method

        # percentage of the best in elite group, or int, the number of best elite
        elite_best = ga_config.get("elite_best", 0.1)

        # percentage of the worst in elite group, or int, the number of worst elite
        elite_worst = ga_config.get("elite_worst", 0.3)

        # "BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", "SingleGA"
        sel_model = ga_config.get("model_selection", "BaseGA")

        # Generate initial solution for inputs
        init_vals = self._generate_initial_solutions(self.init_solution, pop_size)

        if sel_model not in ["BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", "SingleGA"]:
            print("Error: sel_model must be one of the following: "
                  "'BaseGA', 'EliteSingleGA', 'EliteMultiGA', 'MultiGA', 'SingleGA'.")
            print("Defaulting to 'BaseGA'.")
            sel_model = "BaseGA"

        if sel_model == "BaseGA":
            model_ga = GA.BaseGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm, **kwargs)
        elif sel_model == "EliteSingleGA":
            model_ga = GA.EliteSingleGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                        selection=selection,
                                        k_way=k_way,
                                        crossover=crossover,
                                        mutation=mutation,
                                        elite_best=elite_best,
                                        elite_worst=elite_worst, **kwargs)
        elif sel_model == "EliteMultiGA":
            model_ga = GA.EliteMultiGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                       selection=selection,
                                       k_way=k_way,
                                       crossover=crossover,
                                       mutation=mutation,
                                       elite_best=elite_best,
                                       elite_worst=elite_worst, **kwargs)
        elif sel_model == "MultiGA":
            model_ga = GA.MultiGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                  selection=selection,
                                  k_way=k_way,
                                  crossover=crossover,
                                  mutation=mutation, **kwargs)
        elif sel_model == "SingleGA":
            model_ga = GA.SingleGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                   selection=selection,
                                   k_way=k_way,
                                   crossover=crossover,
                                   mutation=mutation, **kwargs)

        # solve the problem
        g_best = model_ga.solve(self.problem_dict, termination=self.term_dict, starting_solutions=init_vals)

        # update files with the best solution
        fitness_func(g_best.solution, scenario_config=self.scenario_config, error_func="rmse")

        return (g_best, model_ga)

    def run_SA(self, **kwargs):
        """Run Simulated Annealing (SA) for behavior optimization.

        Note:
            1. The `model_selection` parameter allows you to choose different types of SA models. Default is "OriginalSA".
                Options include "OriginalSA", "GaussianSA", "SwarmSA".

            SA parameters:
                epoch (int): iterations. Defaults to 1000.
                pop_size (int): population size. Defaults to 2.
                temp_init (float): initial temperature. Defaults to 100.
                cooling_rate (float): Defaults to 0.99.
                scale (float): the change scale of initialization. Defaults to 0.1.
                model_selection (str): select diff. Defaults to "OriginalSA". Options: "OriginalSA", "GaussianSA", "SwarmSA".

        See Also:
            https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.physics_based.html#module-mealpy.physics_based.SA

        Warning:
            You can change the input parameters only from input_config.yaml file.

        Args:
            kwargs: additional keyword arguments for specific SA models (Navigate to See Also).
        """

        if (sa_config := self.behavior_cfg.get("sa_config")) is None:
            raise ValueError("sa_config is not provided in Calibration/turn_inflow setting in yaml file.")

        epoch = sa_config.get("epoch", 1000)  # max iterations
        pop_size = sa_config.get("pop_size", 2)  # population size
        temp_init = sa_config.get("temp_init", 100)  # initial temperature
        cooling_rate = sa_config.get("cooling_rate", 0.891)  # cooling rate
        scale = sa_config.get("scale", 0.1)  # scale of the change
        sel_model = sa_config.get("model_selection", "OriginalSA")  # "OriginalSA", "GaussianSA", "SwarmSA"

        # Generate initial solution for inputs
        init_vals = self._generate_initial_solutions(self.init_solution, pop_size)

        if sel_model not in ["OriginalSA", "GaussianSA", "SwarmSA"]:
            print("Error: sel_model must be one of the following: "
                  "'OriginalSA', 'GaussianSA', 'SwarmSA'.")
            print("Defaulting to 'OriginalSA'.")
            sel_model = "OriginalSA"

        if sel_model == "OriginalSA":
            model_sa = SA.OriginalSA(epoch=epoch,
                                     pop_size=pop_size,
                                     temp_init=temp_init,
                                     cooling_rate=cooling_rate,
                                     step_size=scale,
                                     **kwargs)
        elif sel_model == "GaussianSA":
            model_sa = SA.GaussianSA(epoch=epoch,
                                     pop_size=pop_size,
                                     temp_init=temp_init,
                                     cooling_rate=cooling_rate,
                                     scale=scale,
                                     **kwargs)
        elif sel_model == "SwarmSA":
            model_sa = SA.SwarmSA(epoch=epoch,
                                  pop_size=pop_size,
                                  max_sub_iter=5,
                                  t0=temp_init,
                                  t1=1,
                                  move_count=5,
                                  mutation_rate=0.1,
                                  mutation_step_size=0.1,
                                  mutation_step_size_damp=cooling_rate,
                                  **kwargs)

        g_best = model_sa.solve(self.problem_dict, termination=self.term_dict, starting_solutions=init_vals)

        # update files with the best solution
        fitness_func(g_best.solution, scenario_config=self.scenario_config, error_func="rmse")

        return (g_best, model_sa)

    def run_TS(self, **kwargs):
        """Run Tabu Search (TS) for behavior optimization.

        See Also:
            https://github.com/thieu1995/mealpy/blob/master/mealpy/math_based/TS.py

        Warning:
            You can change the input parameters only from input_config.yaml file.

        Notes:
            TS parameters:
                epoch (int): max iterations. Defaults to 1000.
                pop_size (int): population size. Defaults to 2.
                tabu_size (int): maximum size of tabu list. Defaults to 10.
                neighbour_size (int): size of the neighborhood for generating candidate solutions. Defaults to 10.
                perturbation_scale (float): scale of perturbation for generating candidate solutions. Defaults to 0.05.

        Args:
            kwargs: additional keyword arguments for specific TS models (Navigate to See Also).
        """
        if (ts_config := self.behavior_cfg.get("ts_config")) is None:
            raise ValueError("ts_config is not provided in Calibration/turn_inflow setting in yaml file.")

        epoch = ts_config.get("epoch", 1000)  # max iterations
        pop_size = ts_config.get("pop_size", 2)  # population size
        tabu_size = ts_config.get("tabu_size", 10)  # maximum size of tabu list

        # size of the neighborhood for generating candidate solutions
        neighbour_size = ts_config.get("neighbour_size", 10)

        # scale of perturbation for generating candidate solutions
        perturbation_scale = ts_config.get("perturbation_scale", 0.05)

        # Generate initial solution for inputs
        init_vals = self._generate_initial_solutions(self.init_solution, pop_size)

        model_ts = TS.OriginalTS(epoch=epoch,
                                 pop_size=pop_size,
                                 tabu_size=tabu_size,
                                 neighbour_size=neighbour_size,
                                 perturbation_scale=perturbation_scale,
                                 **kwargs)
        # not print out log to console
        self.problem_dict["log_to"] = "None"
        g_best = model_ts.solve(self.problem_dict, termination=self.term_dict, starting_solutions=init_vals)

        # update files with the best solution
        fitness_func(g_best.solution, scenario_config=self.scenario_config, error_func="rmse")

        return (g_best, model_ts)


if __name__ == "__main__":

    scenario_config = {
        "input_dir": r"C:\Users\xh8\ornl_work\github_workspace\Real-Twin-Dev\realtwin\func_lib\_f_calibration\algo_sumo_\input_dir_dummy",
        "network_name": "chatt",
        "sim_name": "chatt.sumocfg",
        "sim_start_time": 28800,
        "sim_end_time": 32400,
        "path_turn": "chatt.turn.xml",
        "path_inflow": "chatt.flow.xml",
        "path_summary": "summary.xlsx",
        "path_edge": 'EdgeData.xml',
        "calibration_target": {'GEH': 5, 'GEHPercent': 0.85},
        "calibration_interval": 60,
        "demand_interval": 15,
        "EB_tt": 240,
        "WB_tt": 180,
        "EB_edge_list": ["-312", "-293", "-297", "-288", "-286",
                         "-302", "-3221", "-322", "-313", "-284",
                         "-328", "-304"],
        "WB_edge_list": ["-2801", "-280", "-307", "-327", "-281",
                         "-315", "-321", "-300", "-2851", "-285",
                         "-290", "-298", "-295"]
    }

    behavior_config = {"initial_params": {"min_gap": 2.5,        # minimum gap in meters
                                          "acceleration": 2.6,  # max acceleration in m/s^2
                                          "deceleration": 4.5,  # max deceleration in m/s^2
                                          "sigma": 0.5,          # driver imperfection
                                          "tau": 1.00,            # desired headway
                                          "emergencyDecel": 9.0},   # emergency deceleration
                       "params_ranges": {"min_gap": (1.0, 3.0),
                                         "acceleration": (2.5, 3),
                                         "deceleration": (4, 5.3),
                                         "sigma": (0, 1),
                                         "tau": (0.25, 1.25),
                                         "emergencyDecel": (5.0, 9.3)},
                       "max_epoch": 500,
                       "max_fe": 10000,
                       "max_time": 3600,
                       "max_early_stop": 20,

                       "ga_config": {"epoch": 1000,
                                     "pop_size": 30,
                                     "pc": 0.95,
                                     "pm": 0.1,
                                     "selection": "roulette",
                                     "k_way": 0.2,
                                     "crossover": "uniform",
                                     "mutation": "swap",
                                     "elite_best": 0.1,
                                     "elite_worst": 0.3,
                                     "model_selection": "BaseGA",
                                     },
                       "sa_config": {"epoch": 1000,
                                     "temp_init": 100,
                                     "cooling_rate": 0.891,
                                     "scale": 0.1,
                                     "model_selection": "OriginalSA",
                                     },
                       "ts_config": {"epoch": 1000,
                                     "tabu_size": 10,
                                     "neighbour_size": 10,
                                     "perturbation_scale": 0.05,
                                     },
                       }

    opt = BehaviorCali(scenario_config=scenario_config,
                       behavior_config=behavior_config, verbose=True)

    # Run Genetic Algorithm
    # g_best = opt.run_GA()

    # Run Simulated Annealing
    # g_best = opt.run_SA()

    # Run Tabu Search
    g_best = opt.run_TS()



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host Traffic element of Abstract scenario
'''


class Traffic:
    '''The traffic class to host the Traffic element of Abstract scenario'''
    def __init__(self):
        self.AllTrafficTypes = ['Volume', 'TurningRatio', 'Trajectory']
        self.Volume = None
        self.TurningRatio = None
        self.Trajectory = {}
        self.VolumeLookupTable = None

    def isEmpty(self):
        """Check if the Traffic element is empty"""
        pass




##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Email: realtwin@ornl.gov                                                   #
##############################################################################

"""Control of module imports for the RealTwin package."""

import sys
from realtwin._realtwin import RealTwin
from realtwin.func_lib._a_install_simulator.check_sim_env import (is_sumo_installed,
                                                                  is_vissim_installed,
                                                                  is_aimsun_installed)
from realtwin.func_lib import (install_sumo,
                               SimPrep,
                               BehaviorCali,
                               TurnInflowCali,
                               cali_sumo,
                               cali_aimsun,
                               cali_vissim,
                               load_input_configs)

from realtwin.util_lib.create_venv import venv_create, venv_delete
from realtwin.util_lib.create_config import prepare_realtwin_configs
from realtwin.util_lib import get_bbox_from_vertices

# Autonomous vehicle simulation
from realtwin.autonomous_veh import SimAV, prepare_av_configs, load_av_configs


__version__ = '0.1.1'

"The minimum required Python version for RealTwin is 3.10"

__all__ = [
    'RealTwin',
    'prepare_realtwin_configs',
    "load_input_configs",

    # util_lib
    'venv_create', 'venv_delete',
    'is_sumo_installed', 'is_vissim_installed', 'is_aimsun_installed',
    "get_bbox_from_vertices",

    # func_lib
    'install_sumo', 'SimPrep', 'BehaviorCali', 'TurnInflowCali', 'cali_sumo', 'cali_aimsun', 'cali_vissim',

    # autonomous_veh
    'SimAV',
    'prepare_av_configs',
    'load_av_configs',
]


def check_python_version(min_version: str = "3.10") -> tuple:
    """ Check if the current Python version meets the minimum requirement."""
    # Split the version string and convert to tuple of integers
    # version_tuple = tuple(map(int, sys.version.split()[0].split('.')))
    version_tuple = tuple(int(val) for val in sys.version.split()[0].split('.'))

    # Check if the version is greater than or equal to the minimum version required
    major, minor = min_version.split(".")
    try:
        if version_tuple < (int(major), int(minor)):
            raise EnvironmentError(f"Python version {min_version} or higher is required.")
    except Exception:
        print(f"pyufunc supports Python {min_version} or higher.")
    return version_tuple


check_python_version(min_version="3.10")



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Utility functions for calibration of turn ratios and inflow counts in SUMO simulations."""

import os
import sys
import xml.etree.ElementTree as ET
from pathlib import Path
import shutil
import subprocess
import warnings
import pandas as pd
import numpy as np
import pyufunc as pf

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
    sys.path = list(set(sys.path))  # remove duplicates
else:
    warnings.warn("Environment variable 'SUMO_HOME' is not set. "
                  "please declare environment variable 'SUMO_HOME'")
    # sys.exit("please declare environment variable 'SUMO_HOME'")
import traci


def update_turn_flow_from_solution(initial_solution: np.array,
                                   TurnDf: pd.DataFrame,
                                   TurnToCalibrate: pd.DataFrame,
                                   InflowDf: pd.DataFrame,
                                   InflowEdgeToCalibrate: pd.DataFrame,
                                   cali_interval: int,
                                   demand_interval: int) -> tuple:
    """assign the new turn ratios and inflow counts to the given dataframes

    Args:
        initial_solution (np.array): the initial solution from the genetic algorithm
        TurnDf (pd.DataFrame): the turn dataframe from turn.xlsx
        TurnToCalibrate (pd.DataFrame): the turn dataframe from TurnToCalibrate.xlsx
        InflowDf (pd.DataFrame): the inflow dataframe from inflow.xlsx
        InflowEdgeToCalibrate (list): the list of edges to calibrate inflow
        cali_interval (int): the calibration interval
        demand_interval (int): the demand interval

    Returns:
        tuple: the updated turn and inflow dataframes
    """

    i = 0  # index into initial_solution
    # Loop through each group of (JunctionID_OpenDrive, Numbering)
    grouped = TurnToCalibrate.groupby(["JunctionID_OpenDrive", "Numbering"])
    for _, group in grouped:
        # indices of calibrated / non-calibrated rows in this group
        cal_idx = group.index[group["Calibration variable?"] == 1].tolist()
        ncal_idx = group.index[group["Calibration variable?"] == 0].tolist()

        # take values for calibrated ones from the solution vector
        n_cal = len(cal_idx)
        if n_cal > 0:
            vals = np.array(initial_solution[i:i+n_cal], dtype=float)
            vals = np.clip(vals, 0, None)                # ensure >= 0
            s = vals.sum()

            if s > 1.0:                                  # cap at 1 (preserve ratios)
                vals = vals / s                          # sum becomes 1
                s = 1.0

            # assign calibrated ratios
            for k, idx in enumerate(cal_idx):
                x = group.loc[idx, "OpenDriveFromID"]
                y = group.loc[idx, "OpenDriveToID"]
                TurnDf.loc[
                    (TurnDf["OpenDriveFromID"] == x) & (TurnDf["OpenDriveToID"] == y),
                    "TurnRatio"
                ] = vals[k]

            i += n_cal
        else:
            s = 0.0  # no calibrated vars in this group

        # distribute remainder to non-calibrated based on existing weights (fallback = equal)
        remainder = max(1.0 - s, 0.0)

        if len(ncal_idx) > 0 and remainder > 0:
            # read current TurnRatio as weights
            weights = TurnDf.loc[ncal_idx, "TurnRatio"].astype(float).to_numpy()
            wsum = weights.sum()
            if wsum > 0:
                weights = weights / wsum
            else:
                weights = np.ones(len(ncal_idx)) / len(ncal_idx)

            # assign weighted remainder
            for k, idx in enumerate(ncal_idx):
                x = group.loc[idx, "OpenDriveFromID"]
                y = group.loc[idx, "OpenDriveToID"]
                TurnDf.loc[
                    (TurnDf["OpenDriveFromID"] == x) & (TurnDf["OpenDriveToID"] == y),
                    "TurnRatio"
                ] = remainder * weights[k]
        elif len(ncal_idx) > 0 and remainder == 0:
            for idx in ncal_idx:
                x = group.loc[idx, "OpenDriveFromID"]
                y = group.loc[idx, "OpenDriveToID"]
                TurnDf.loc[
                    (TurnDf["OpenDriveFromID"] == x) & (TurnDf["OpenDriveToID"] == y),
                    "TurnRatio"
                ] = 0.0

    # Inflow calibration assignments based on InflowEdgeToCalibrate
    for edge in InflowEdgeToCalibrate:
        InflowDf.loc[InflowDf['OpenDriveFromID'] == edge, 'Count'] = initial_solution[i] / cali_interval * demand_interval
        i += 1

    return (TurnDf, InflowDf)


def run_jtrrouter_to_create_rou_xml(network_name: str, path_net: str,
                                    TurnDf: pd.DataFrame, InflowDf: pd.DataFrame,
                                    path_rou: str, sim_start_time: float,
                                    sim_end_time: float, verbose: bool = False) -> None:
    """Runs jtrrouter to generate a route file from flow and network files in SUMO.

    Args:
        network_name (str): The name of the network.
        path_net (str): The path to the network file.
        TurnDf (pd.DataFrame): DataFrame containing turn ratio data.
        InflowDf (pd.DataFrame): DataFrame containing inflow count data.
        path_rou (str): The path to the output route file.
        sim_start_time (float): The start time of the simulation.
        sim_end_time (float): The end time of the simulation.
        verbose (bool): If True, print additional information. Defaults to False.
    """

    dir_rou = Path(path_rou).parent
    dir_net = Path(path_net).parent
    path_turn = pf.path2linux(dir_rou / f"{network_name}.turn.xml")
    path_flow = pf.path2linux(dir_rou / f"{network_name}.flow.xml")

    TurnDf['IntervalStart'] = TurnDf['IntervalStart'].astype(float)
    TurnDf['IntervalEnd'] = TurnDf['IntervalEnd'].astype(float)
    TurnDf = TurnDf[(TurnDf['IntervalStart'] >= sim_start_time) & (TurnDf['IntervalEnd'] <= sim_end_time)]
    turns = ET.Element('turns')
    # Create the 'interval' element
    IntervalSet = TurnDf[['IntervalStart', 'IntervalEnd']].drop_duplicates().reset_index(drop=True)
    for _, IntervalData in IntervalSet.iterrows():
        Interval = ET.SubElement(turns, 'interval')
        Interval.set('begin', str(IntervalData['IntervalStart']))
        Interval.set('end', str(IntervalData['IntervalEnd']))

        TurnDfSubset = TurnDf[(TurnDf['IntervalStart'] == IntervalData['IntervalStart']) & (
            TurnDf['IntervalEnd'] == IntervalData['IntervalEnd'])]
        TurnDictSubset = TurnDfSubset.to_dict(orient='records')
        for TurnData in TurnDictSubset:
            edge_relation = ET.SubElement(Interval, 'edgeRelation')
            # edge_relation.set('from', f"-{TurnData['OpenDriveFromID']}")
            # edge_relation.set('to', f"-{TurnData['OpenDriveToID']}")
            edge_relation.set('from', "-" + str(TurnData['OpenDriveFromID']))
            edge_relation.set('to',   "-" + str(TurnData['OpenDriveToID']))  
            edge_relation.set('probability', str(TurnData['TurnRatio']))
    # <edgeRelation from="" probability="" to=""/>
    TreeTurn = ET.ElementTree(turns)
    # Write the XML to the file
    TreeTurn.write(path_turn, encoding='utf-8', xml_declaration=True)

    # Create the .flow.xml
    InflowDf['IntervalStart'] = InflowDf['IntervalStart'].astype(float)
    InflowDf['IntervalEnd'] = InflowDf['IntervalEnd'].astype(float)
    InflowDf = InflowDf[(InflowDf['IntervalStart'] >= sim_start_time) & (InflowDf['IntervalEnd'] <= sim_end_time)]
    routes = ET.Element('routes')
    vtype = ET.SubElement(routes, 'vType')
    vtype.set('id', 'car')
    vtype.set('type', 'passenger')
    InflowDict = InflowDf.to_dict(orient='records')
    FlowID = 0
    for InflowData in InflowDict:
        FlowID += 1
        flow = ET.SubElement(routes, 'flow')
        flow.set('id', str(FlowID))
        flow.set('begin', str(InflowData['IntervalStart']))
        flow.set('end', str(InflowData['IntervalEnd']))
        # flow.set('from', f"-{InflowData['OpenDriveFromID']}")
        flow.set('from', "-" + str(InflowData['OpenDriveFromID']))
        flow.set('number', str(int(InflowData['Count'])))
        # flow.set('number', str(int(int(InflowData['Count'])/CablibrationInterval*DemandInterval)))
        flow.set('type', 'car')

    # <flow begin="0.0" end="3600.0" from="" id="" number="" type="car"/>
    TreeInflow = ET.ElementTree(routes)
    # TreeInflow.write('MyNetwork/SUMO/{}.flow.xml'.format(NetworkName), encoding='utf-8', xml_declaration=True)
    TreeInflow.write(path_flow, encoding='utf-8', xml_declaration=True)

    # Define the jtrrouter command with all necessary arguments
    # cmd = [
    #     "jtrrouter",
    #     "-n", path_net,
    #     "-r", path_flow,
    #     "-t", path_turn,
    #     "-o", path_rou,
    #     "--accept-all-destinations",
    #     "--remove-loops True",
    #     "--randomize-flows",
    #     # "--seed","101",
    #     # "--ignore-errors",  # Continue on errors; remove if not desired
    # ]
    # cmd = f'jtrrouter -r "{path_flow}" -t "{path_turn}" -n "{path_net}" --accept-all-destinations --remove-loops True --randomize-flows -o "{path_rou}"'
    cmd = (
        f'jtrrouter '
        f'-r "{path_flow}" '
        f'-t "{path_turn}" '
        f'-n "{path_net}" '
        f'--accept-all-destinations '
        f'--remove-loops True '
        f'--randomize-flows '
        f'-o "{path_rou}"'
    )

    # Execute the command
    try:
        # subprocess.run(cmd, capture_output=True, text=True)
        process = subprocess.Popen(cmd, shell=True)
        # process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        # # Print output line by line in real time
        # for line in process.stdout:
        #     print(line, end='')  # end='' prevents double newlines

        process.wait()
        if verbose:
            print(f"  :Route file generated successfully: {path_rou}")

        shutil.copy(path_rou, dir_net)
    except subprocess.CalledProcessError as e:
        print(f"  :An error occurred while running jtrrouter: {e}")


def run_SUMO_create_EdgeData(sim_name: str, sim_end_time: float) -> bool:
    """run SUMO simulation using traci module

    Args:
        sim_name (str): the name of the simulation, it should be the .sumocfg file
        sim_end_time (float): the end time of the simulation

    Returns:
        bool: True if the simulation is successful
    """

    traci.start(["sumo", "-c", sim_name])
    while traci.simulation.getTime() < sim_end_time:
        traci.simulationStep()
    traci.close()
    return True


def result_analysis_on_EdgeData(Summary_data: pd.DataFrame,
                                path_EdgeData: str,
                                calibration_target: dict,
                                sim_start_time: float,
                                sim_end_time: float) -> tuple:
    """Analyze the result of the simulation and return the flag, mean GEH, and GEH percent

    Args:
        Summary_data (pd.DataFrame): the summary dataframe from summary.xlsx in input dir
        path_EdgeData (str): the path to the EdgeData.xml file in the input dir
        calibration_target (dict): the calibration target from the scenario config, it should contain GEH and GEHPercent
        sim_start_time (float): the start time of the simulation
        sim_end_time (float): the end time of the simulation

    Returns:
        tuple: (flag, mean GEH, geh percent)
    """
    RealSummary = Summary_data[Summary_data["realcount"].notna()]
    RealSummary['IntervalStart'] = pd.to_numeric(RealSummary['IntervalStart'], errors='coerce')
    RealSummary['IntervalEnd']   = pd.to_numeric(RealSummary['IntervalEnd'], errors='coerce')    
    Interval =   (RealSummary['IntervalEnd'].max()-RealSummary['IntervalStart'].min())/3600

    ApproachSummary = RealSummary.groupby(['IntersectionName',
                                           'entrance_sumo',
                                           'Bound']).agg({'realcount': 'sum'}).reset_index()
    tree = ET.parse(path_EdgeData)
    root = tree.getroot()
    edge_data = [
        {
            'id': str(edge.get('id')) if edge.get('id') else None,
            'travel_time': float(edge.get('traveltime')) if edge.get('traveltime') else None,
            'arrived': int(edge.get('arrived')) if edge.get('arrived') else None,
            'departed': int(edge.get('departed')) if edge.get('departed') else None,
            'left': int(edge.get('left')) if edge.get('left') else None,
            'density': float(edge.get('density')) if edge.get('density') else None,
            'speed': float(edge.get('speed')) if edge.get('speed') else None
        }
        for interval in root.findall('.//interval')
        for edge in interval.findall('edge')
    ]
    EdgeData = pd.DataFrame(edge_data)

    EdgeData = EdgeData.astype({'id': str,
                                'travel_time': float,
                                'arrived': int,
                                'departed': int,
                                "left": int,
                                'density': float,
                                'speed': float})
    ApproachSummary['entrance_sumo'] = ApproachSummary['entrance_sumo'].astype(str)
    EdgeData['id'] = EdgeData['id'].astype(str)
    ApproachSummary = pd.merge(ApproachSummary, EdgeData, left_on='entrance_sumo', right_on='id')
    ApproachSummary.rename(columns={'left': 'count'}, inplace=True)
    ApproachSummary.drop(columns=['id'], inplace=True)
    ApproachSummary['flow'] = ApproachSummary['count'] /Interval
    ApproachSummary['realflow'] = ApproachSummary['realcount'] /Interval

    ApproachSummary['GEH'] = np.sqrt(2 * (ApproachSummary['flow'] - ApproachSummary['realflow'])
                                     ** 2 / (ApproachSummary['flow'] + ApproachSummary['realflow']))
    MeanGEH = ApproachSummary['GEH'].mean()
    GEHPercent = (ApproachSummary['GEH'] < calibration_target['GEH']).mean()
    flag = 1
    # more than 85% of links have a GEH <= 5
    if GEHPercent < calibration_target['GEHPercent']:
        flag = 0

    BadVolume = ApproachSummary[ApproachSummary['count'] < 0]
    # within 100 vph for volumes < 700
    df1 = ApproachSummary[ApproachSummary['realflow'] < 700]
    if sum((df1['realflow'] - df1['flow']).abs() > 100) > 0:
        flag = 0
        BadVolume = pd.concat([BadVolume, (df1[(df1['realflow'] - df1['flow']).abs() > 100])])

    # within 15%  for volumes  700-2700
    df2 = ApproachSummary[(ApproachSummary['realflow'] >= 700) & (ApproachSummary['realflow'] <= 2700)]
    if sum(((df2['realflow'] - df2['flow']) / -df2['realflow']).abs() > 0.15) > 0:
        flag = 0
        BadVolume = pd.concat([BadVolume, (df2[((df2['realflow'] - df2['flow']) / -df2['realflow']).abs() > 0.15])])

    # within 400 vph for volumes > 2700
    df3 = ApproachSummary[ApproachSummary['realflow'] > 2700]
    if sum((df3['realflow'] - df3['flow']).abs() > 400) > 0:
        flag = 0
        BadVolume = pd.concat([BadVolume, (df3[(df3['realflow'] - df3['flow']).abs() > 400])])
    # ApproachSummary.to_excel("ApproachSummary.xlsx", sheet_name="Summary", index=False)
    return (flag, MeanGEH, GEHPercent)


def time_to_seconds(time_str):
    """ Convert a time string in the format 'HH:MM' to seconds."""
    hour, minute = [int(x) for x in time_str.split(':')]
    return (hour * 3600) + (minute * 60)


def read_MatchupTable(path_matchup_table: str) -> pd.DataFrame:
    """ Load the matchup table from user input

    Args:
        path_matchup_table (str): the path to user updated matchup table

    Returns:
        pd.DataFrame: the loaded matchup table dataframe.
    """

    MatchupTable_UserInput = pd.read_excel(path_matchup_table, skiprows=1, dtype=str)

    # Forward fill missing values in merged columns
    merged_columns = ["JunctionID_OpenDrive",
                      "IntersectionName_GridSmart", "File_Synchro", "Need calibration?"]
    MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill()
    return MatchupTable_UserInput


def generate_turn_demand_cali(*, path_matchup_table: str | pd.DataFrame,
                              traffic_dir: str) -> list[pd.DataFrame]:
    """ Generate turn demand from user input lookup table and Synchro UTDF files.

    Args:
        path_matchup_table (str): Path to the matchup table with user input.
        traffic_dir (str): Directory where demand files are located.
            Defaults to "Traffic".

    See Also:
        traffic_dir: check sample demand files in datasets/Traffic directory

    Example:
        >>> path_matchup_table = "./MatchupTable_OpenDrive_with user input.xlsx"
        >>> TurnDf, IDRef = generate_turn_demand(path_matchup_table, output_dir="./Output", traffic_dir="Traffic")

    Returns:
        list[pd.DataFrame]: A list containing two DataFrames:
            - TurnDf: DataFrame with turn demand data.
            - IDRef: DataFrame with reference IDs for OpenDrive turns (demand lookup table).
    """

    if isinstance(path_matchup_table, str):
        # Load the MatchupTable_OpenDrive_withsignal.xlsx file, skipping the first row for correct headers
        MatchupTable_UserInput = pd.read_excel(path_matchup_table, skiprows=1, dtype=str)

        # Forward fill missing values in merged columns
        merged_columns = ["JunctionID_OpenDrive", "IntersectionName_GridSmart", "File_Synchro", "Need calibration?"]
        MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill()
    elif isinstance(path_matchup_table, pd.DataFrame):
        MatchupTable_UserInput = path_matchup_table
    else:
        raise Exception("  : Invalid path_matchup_table, it should be in string path or pandas DataFrame.")

    turn_values = ["NBR", "NBT", "NBL", "NBU", "EBR", "EBT", "EBL", "EBU",
                   "SBR", "SBT", "SBL", "SBU", "WBR", "WBT", "WBL", "WBU"]
    TurnDf_list = []
    IDRef_list = []

    # Process each unique JunctionID_OpenDrive where File_GridSmart has input
    for junction_id in MatchupTable_UserInput["JunctionID_OpenDrive"].unique():
        subset = MatchupTable_UserInput[MatchupTable_UserInput["JunctionID_OpenDrive"] == junction_id]

        # Get the file path from File_GridSmart if available
        file_name = subset["File_GridSmart"].dropna().iloc[0] if not subset["File_GridSmart"].isna().all() else None

        if file_name:
            # Retrieve the IntersectionName_GridSmart
            intersection_name = subset["IntersectionName_GridSmart"].dropna().iloc[0] if not subset["IntersectionName_GridSmart"].isna().all() else "Unknown"
            # Create df_lookup with predefined Turn values
            df_lookup = pd.DataFrame({"Turn": turn_values})
            # Assign IntersectionName
            df_lookup["IntersectionName"] = intersection_name
            # Initialize OpenDriveFromID and OpenDriveToID as empty strings
            df_lookup["OpenDriveFromID"] = ""
            df_lookup["OpenDriveToID"] = ""
            # Map values from subset based on Turn_GridSmart
            for idx, row in df_lookup.iterrows():
                turn = row["Turn"]
                match = subset[subset["Turn_GridSmart"] == turn]
                if not match.empty:
                    df_lookup.at[idx, "OpenDriveFromID"] = match["FromRoadID_OpenDrive"].values[0] if not match["FromRoadID_OpenDrive"].isna().all() else ""
                    df_lookup.at[idx, "OpenDriveToID"] = match["ToRoadID_OpenDrive"].values[0] if not match["ToRoadID_OpenDrive"].isna().all() else ""
            # Append to the list
            IDRef_list.append(df_lookup)

        # Get the file path from File_GridSmart if available
        file_name = subset["File_GridSmart"].dropna().iloc[0] if not subset["File_GridSmart"].isna().all() else None

        if file_name:
            if '.' not in Path(file_name).name:
                file_name = file_name + '.xls'
            gs_file_path = os.path.join(traffic_dir, file_name)
            # gs_file_path = f"GridSmart/{file_name}"

            # Check if the file exists before processing
            if os.path.exists(gs_file_path):
                # Load the Excel file
                df = pd.read_excel(gs_file_path, header=None)  # Read without predefined headers

                # Find the first row that contains a time value in the first column
                time_row_index = df[df[0].astype(str).str.match(r'^\d{1,2}:\d{2}$', na=False)].index.min()

                # Determine the starting row for data extraction
                start_row = time_row_index - 2 if pd.notna(time_row_index) else None

                if start_row is not None:
                    # Read the file again with proper headers starting from the determined row
                    df_data = pd.read_excel(gs_file_path, header=[start_row, start_row + 1])

                    # Fill merged cells in the first row
                    df_data.columns = df_data.columns.to_frame().fillna(method="ffill").agg("".join, axis=1)

                    # Remove spaces from column names
                    df_data.columns = [col.replace(" ", "") for col in df_data.columns]

                    # Rename the first column to "Time"
                    df_data.rename(columns={df_data.columns[0]: "Time"}, inplace=True)

                    # Drop fully empty columns
                    df_data.dropna(axis=1, how='all', inplace=True)

                    # Drop data of "Total"
                    df_data = df_data[df_data["Time"] != "Total"]

                    # Convert data columns to numeric
                    for col in df_data.columns[1:]:  # Excluding 'Time' column
                        df_data[col] = pd.to_numeric(df_data[col], errors='coerce').fillna(0).astype(int)

                    # Remove 'Unassigned' columns
                    df_data = df_data.loc[:, ~df_data.columns.str.contains(r'Unassigned', na=False)]

                    # Standardize column names for directions
                    df_data.columns = [col.replace("Northbound", "NB")
                                       .replace("Southbound", "SB")
                                       .replace("Westbound", "WB")
                                       .replace("Eastbound", "EB") for col in df_data.columns]

                    # Define expected columns and ensure all are present
                    expected_columns = ["IntersectionName", "Time", "NBR", "NBT", "NBL", "NBU",
                                        "EBR", "EBT", "EBL", "EBU", "SBR", "SBT", "SBL", "SBU",
                                        "WBR", "WBT", "WBL", "WBU"]
                    df_data = df_data.reindex(columns=expected_columns, fill_value="")

                    # Fill IntersectionName using IntersectionName_GridSmart from MatchupTable_UserInput
                    intersection_name = subset["IntersectionName_GridSmart"].dropna().iloc[0] if not subset["IntersectionName_GridSmart"].isna().all() else "Unknown"
                    df_data["IntersectionName"] = intersection_name

                    # Append processed data to list
                    TurnDf_list.append(df_data)

    TurnDf = pd.concat(TurnDf_list, ignore_index=True) if TurnDf_list else pd.DataFrame()

    IDRef = pd.concat(IDRef_list, ignore_index=True) if IDRef_list else pd.DataFrame()
    IDRef = IDRef[["IntersectionName", "Turn", "OpenDriveFromID", "OpenDriveToID"]]

    # replace "" to numpy.nan
    # TurnDf = TurnDf.replace("", np.nan)
    # IDRef = IDRef.replace("", np.nan)
    IDRef = IDRef.dropna(subset=['OpenDriveFromID', 'OpenDriveToID'])

    # drop '' in column OpenDriveFromID and OpenDriveToID
    IDRef = IDRef[IDRef["OpenDriveFromID"].astype(str) != ""]
    IDRef = IDRef[IDRef["OpenDriveToID"].astype(str) != ""]

    return [TurnDf, IDRef]


def generate_inflow(path_net: str,
                    MatchupTable_UserInput: pd.DataFrame,
                    TurnDf: pd.DataFrame,
                    IDRef: pd.DataFrame,
                    sim_begin: int = 28800,
                    sim_end: int = 32400,):
    """ Generate inflow data for calibration."""
    # Apply the conversion function to the 'Time' column and create a new 'Seconds' column
    TurnDf['IntervalStart'] = TurnDf['Time'].apply(time_to_seconds)
    TurnDf['IntervalEnd'] = TurnDf['IntervalStart'] + 15 * 60
    TurnDf = TurnDf.drop('Time', axis=1)

    # Reshape the DataFrame to the long format
    TurnDfTemp = TurnDf.melt(id_vars=['IntersectionName', 'IntervalStart', 'IntervalEnd'],
                             var_name='Turn',
                             value_name='Count')
    # Sort the DataFrame by IntersectionName and Turn columns
    TurnDfTemp.sort_values(['IntersectionName', 'IntervalStart', 'IntervalEnd'], inplace=True)
    # Reset the index
    TurnDfTemp.reset_index(drop=True, inplace=True)
    Count = TurnDfTemp.copy()

    IDRef = IDRef.dropna(subset=['OpenDriveFromID', 'OpenDriveToID'])
    IDRef = IDRef[(IDRef['OpenDriveFromID'] != '') & (IDRef['OpenDriveToID'] != '')]
    # IDRef = IDRef.astype({'OpenDriveFromID': int, 'OpenDriveToID': int})
    IDRef = IDRef.astype(str)

    MergedDf1 = pd.merge(Count, IDRef, on=['IntersectionName', 'Turn'], how='left')
    Count['OpenDriveFromID'] = MergedDf1['OpenDriveFromID']
    Count['OpenDriveToID'] = MergedDf1['OpenDriveToID']
    Count['Count'] = Count['Count'].replace('', 0)  # Replace empty strings with 0
    Count['Count'] = Count['Count'].fillna(0)       # Fill NaNs with 0
    Count['Count'] = pd.to_numeric(Count['Count'], errors='coerce').fillna(0).astype(int)  # Ensure it's int
    Count = Count.groupby(['IntervalStart', 'IntervalEnd', 'IntersectionName', 'OpenDriveFromID'],
                          as_index=False)['Count'].sum()
    Count = Count.dropna(subset=['OpenDriveFromID'])
    Count = Count.astype({'Count': int})

    tree = ET.parse(path_net)
    root = tree.getroot()
    # DeadEndJunction = set()
    # for junction in root.findall('junction'):
    #     if junction.attrib.get('type') == 'dead_end':
    #         DeadEndJunction.add(junction.attrib['id'])

    conn_from_map = {}
    for conn in root.findall("connection"):
        from_edge = conn.attrib.get("from")
        direction = conn.attrib.get("dir")
        if from_edge:
            if from_edge not in conn_from_map:
                conn_from_map[from_edge] = []
            conn_from_map[from_edge].append(direction)

    DeadEndJunction = set()
    for junction in root.findall("junction"):
        jid = junction.attrib['id']
        jtype = junction.attrib.get('type')
        incLanes = junction.attrib.get('incLanes', '')
        if jtype == 'dead_end':
            DeadEndJunction.add(jid)
            continue

        if jtype == 'internal':
            continue

        if incLanes == '':
            DeadEndJunction.add(jid)
            continue

        lane_ids = incLanes.strip().split()
        edge_ids = set(lane.split('_')[0] for lane in lane_ids)
        match = True
        for edge_id in edge_ids:
            if edge_id in conn_from_map:
                if any(d != 't' for d in conn_from_map[edge_id]):
                    match = False
                    break
        if match:
            DeadEndJunction.add(jid)

    InflowRoad = []
    SingleRoad = []
    for edge in root.findall('edge'):
        edge_from = edge.attrib.get('from')
        edge_to = edge.attrib.get('to')
        edge_id = edge.attrib.get('id')
        if edge_from in DeadEndJunction:
            InflowRoad.append(edge_id.lstrip('-'))
        if edge_from in DeadEndJunction and edge_to in DeadEndJunction:
            SingleRoad.append(edge_id.lstrip('-'))
    FromRoadID_Sumo = ["-" + str(x) for x in MatchupTable_UserInput["FromRoadID_OpenDrive"].dropna().unique()]
    Lookup_InflowEdge = pd.DataFrame(columns=["FromRoadID_Sumo", "InflowID_Sumo"])
    edges = {edge.get("id"): edge for edge in root.findall("edge")}
    junctions = {junc.get("id"): junc for junc in root.findall("junction")}
    for from_id in FromRoadID_Sumo:
        current_id = from_id
        visited = set()
        while current_id in edges and current_id not in visited:
            visited.add(current_id)
            edge = edges[current_id]
            from_junction_id = edge.get("from")
            junction = junctions.get(from_junction_id)
            if junction is None:
                break
            # if junction.get("type") == "dead_end":
            if junction.get("id") in DeadEndJunction:
                Lookup_InflowEdge.loc[len(Lookup_InflowEdge)] = [
                    from_id, current_id]
                break
            incLanes = junction.get("incLanes")
            if not incLanes:
                break
            lane_ids = [lane.split("_")[0] for lane in incLanes.split()]
            lane_ids = list(set(lane_ids))  # Unique
            lane_ids = [x for x in lane_ids if x]  # Remove empty strings
            if len(lane_ids) != 1:
                break
            next_id = lane_ids[0]
            if next_id not in edges:
                break
            current_id = next_id

    Lookup_InflowEdge["FromRoadID_Sumo_stripped"] = Lookup_InflowEdge["FromRoadID_Sumo"].str.lstrip("-")
    Count = Count.merge(
        Lookup_InflowEdge,
        left_on="OpenDriveFromID",
        right_on="FromRoadID_Sumo_stripped",
        how="left"
    )
    Count["InflowID_Sumo_stripped"] = Count["InflowID_Sumo"].str.lstrip("-")
    InflowCount = Count[Count["InflowID_Sumo_stripped"].isin(InflowRoad)].copy()
    InflowCount.loc[InflowCount["InflowID_Sumo_stripped"].notna(),
                    "OpenDriveFromID"] = InflowCount["InflowID_Sumo_stripped"]
    InflowCount.drop(columns=["InflowID_Sumo",
                              "InflowID_Sumo_stripped",
                              "FromRoadID_Sumo",
                              "FromRoadID_Sumo_stripped"],
                     inplace=True)
    Count.drop(columns=["InflowID_Sumo",
                        "InflowID_Sumo_stripped",
                        "FromRoadID_Sumo",
                        "FromRoadID_Sumo_stripped"],
               inplace=True)
    
    # create InflowDf_Calibration for turn and inflow purpose
    InflowDf_Calibration = InflowCount[(InflowCount["IntervalStart"] >= sim_begin) &
                                       (InflowCount["IntervalEnd"] <= sim_end)].copy()
    InflowDf_Calibration["OpenDriveFromID"] = InflowDf_Calibration["OpenDriveFromID"].astype(str)
    InflowDf_Calibration["Count"] = InflowDf_Calibration["Count"].astype(int)
    intervals = InflowDf_Calibration[["IntervalStart", "IntervalEnd"]].drop_duplicates()

    # Inflow edges to calibrate
    MissingInflow = [
        road for road in InflowRoad
        if road not in Count['OpenDriveFromID'].unique() and road not in SingleRoad
    ]
    filtered_missing_inflow = MatchupTable_UserInput[
        (MatchupTable_UserInput["FromRoadID_OpenDrive"].isin(MissingInflow)) &
        (MatchupTable_UserInput["Need calibration?"] == "Y")
    ]
    InflowEdgeToCalibrate = filtered_missing_inflow['FromRoadID_OpenDrive'].dropna().unique().tolist()
    N_InflowVariable = len(InflowEdgeToCalibrate)

    calibration_rows = []
    for _, row in intervals.iterrows():
        for edge in InflowEdgeToCalibrate:
            calibration_rows.append({
                "IntervalStart": row["IntervalStart"],
                "IntervalEnd": row["IntervalEnd"],
                "IntersectionName": f"Calibration for Edge {edge}",
                "OpenDriveFromID": edge,
                "Count": 0})
    calibration_df = pd.DataFrame(calibration_rows)
    InflowDf_Calibration = pd.concat([InflowDf_Calibration, calibration_df], ignore_index=True)
    InflowDf_Calibration.sort_values(by=["IntervalStart", "IntersectionName", "OpenDriveFromID"], inplace=True)
    InflowDf_Calibration.reset_index(drop=True, inplace=True)

    return InflowDf_Calibration, InflowEdgeToCalibrate, N_InflowVariable


def generate_turn_summary(TurnDf: pd.DataFrame, MatchupTable_UserInput: pd.DataFrame, N_InflowVariable: int, sim_begin: int = 28800, sim_end: int = 32400, allow_u_turn: bool = False):
    """ Generate turn summary for calibration."""

    # create TurnDf_Calibration for turn and inflow purpose
    TurnDfTemp = TurnDf.melt(id_vars=['IntersectionName', 'IntervalStart', 'IntervalEnd'],
                             var_name='Turn',
                             value_name='Count')
    TurnDfTemp.sort_values(['IntersectionName', 'IntervalStart', 'IntervalEnd'], inplace=True)
    TurnDfTemp.reset_index(drop=True, inplace=True)
    TurnDfTemp['Bound'] = TurnDfTemp['Turn'].str[0]
    TurnDfTemp['Direction'] = TurnDfTemp['Turn'].str[-1]
    TurnDfTemp['Count'] = TurnDfTemp['Count'].replace('', 0)  # Replace empty strings with 0
    TurnDfTemp['Count'] = TurnDfTemp['Count'].fillna(0)       # Fill NaNs with 0
    TurnDfTemp['Count'] = pd.to_numeric(TurnDfTemp['Count'], errors='coerce').fillna(
        0).astype(int)  # Ensure integer type
    FlowTemp = TurnDfTemp.groupby(['IntervalStart',
                                   'IntervalEnd',
                                   'IntersectionName',
                                   'Bound'],
                                  as_index=False)['Count'].sum()
    Turn = pd.merge(TurnDfTemp, FlowTemp, on=['IntervalStart',
                                              'IntervalEnd',
                                              'IntersectionName',
                                              'Bound'],
                    how='left')
    Turn['realcount'] = Turn['Count_x']
    Turn['TurnRatio'] = Turn['Count_x'] / Turn['Count_y']
    Turn = Turn.drop(['Count_x', 'Count_y'], axis=1)
    Turn.reset_index(drop=True, inplace=True)

    Turn = Turn.merge(
        MatchupTable_UserInput[["IntersectionName_GridSmart",
                                "Turn_GridSmart",
                                "Bearing",
                                "Numbering",
                                "FromRoadID_OpenDrive",
                                "ToRoadID_OpenDrive"]],
        how="left",
        left_on=["IntersectionName", "Turn"],
        right_on=["IntersectionName_GridSmart", "Turn_GridSmart"])

    Turn.drop(columns=["IntersectionName_GridSmart", "Turn_GridSmart"], inplace=True)
    Turn_Calibration = Turn[(Turn["IntervalStart"] >= sim_begin) & (Turn["IntervalEnd"] <= sim_end)].copy()
    Turn_Calibration.dropna(subset=["FromRoadID_OpenDrive", "ToRoadID_OpenDrive", "TurnRatio"], inplace=True)

    TurnDf_Calibration = Turn_Calibration.copy()
    TurnDf_Calibration.rename(columns={"FromRoadID_OpenDrive": "OpenDriveFromID",
                                       "ToRoadID_OpenDrive": "OpenDriveToID"}, inplace=True)

    # Turning ratios to calibrate
    MatchupTable_UserInput["merge_key"] = (
        MatchupTable_UserInput["IntersectionName_GridSmart"].astype(str) + "||" +
        MatchupTable_UserInput["Turn_GridSmart"].astype(str))
    Turn["merge_key"] = Turn["IntersectionName"].astype(str) + "||" + Turn["Turn"].astype(str)
    TurnToCalibrate = MatchupTable_UserInput[
        (~MatchupTable_UserInput["merge_key"].isin(Turn["merge_key"])) &
        (MatchupTable_UserInput["Need calibration?"] == "Y")].copy()
    MatchupTable_UserInput.drop(columns=["merge_key"], inplace=True)
    Turn.drop(columns=["merge_key"], inplace=True)
    TurnToCalibrate = TurnToCalibrate[['JunctionID_OpenDrive',
                                       'Bearing',
                                       'Numbering',
                                       'FromRoadID_OpenDrive',
                                       'ToRoadID_OpenDrive',
                                       'Turn']]
    TurnToCalibrate.rename(columns={
        "FromRoadID_OpenDrive": "OpenDriveFromID",
        "ToRoadID_OpenDrive": "OpenDriveToID"}, inplace=True)

    intervals = TurnDf_Calibration[[
        "IntervalStart", "IntervalEnd"]].drop_duplicates()
    intervals["key"] = 1
    TurnToCalibrate["key"] = 1
    new_rows = pd.merge(intervals, TurnToCalibrate, on="key").drop(columns=["key"])
    TurnDf_Calibration = pd.concat([TurnDf_Calibration, new_rows], ignore_index=True)
    TurnDf_Calibration.sort_values(by=["IntervalStart",
                                       "IntersectionName",
                                       "JunctionID_OpenDrive"],
                                   inplace=True)
    TurnDf_Calibration.reset_index(drop=True, inplace=True)
    TurnToCalibrate.drop(columns=["key"], inplace=True)
    TurnDf_Calibration["TurnRatio"] = TurnDf_Calibration["TurnRatio"].fillna(0)

    # If U-turns are not allowed, drop them
    if not allow_u_turn:
        TurnToCalibrate = TurnToCalibrate[TurnToCalibrate["Turn"] != "Uturn"].copy()

    TurnToCalibrate.sort_values(by=["JunctionID_OpenDrive", "Numbering", "Turn"], inplace=True)
    TurnToCalibrate.reset_index(drop=True, inplace=True)

    TurnToCalibrate["Calibration variable?"] = 1
    last_indices = TurnToCalibrate.groupby(["JunctionID_OpenDrive", "Numbering"]).tail(1).index
    TurnToCalibrate.loc[last_indices, "Calibration variable?"] = 0

    N_TurnVariable = (TurnToCalibrate["Calibration variable?"] == 1).sum()
    N_Variable = N_TurnVariable + N_InflowVariable

    # create RealSummary_Calibration for turn and inflow purpose
    RealSummary_Calibration = Turn_Calibration.copy()
    RealSummary_Calibration["entrance_sumo"] = "-" + RealSummary_Calibration["FromRoadID_OpenDrive"].astype(str)
    RealSummary_Calibration["exit_sumo"] = "-" + RealSummary_Calibration["ToRoadID_OpenDrive"].astype(str)

    return TurnToCalibrate, TurnDf_Calibration, RealSummary_Calibration, N_Variable, N_TurnVariable


if __name__ == "__main__":
    pass
#     path_matchup = "./MatchupTable.xlsx"
#     path_net = "./chatt.net.xml"
#
#     MatchupTable_UserInput = read_MatchupTable(path_matchup_table=path_matchup)
#     TurnDf, IDRef = generate_turn_demand_cali(path_matchup_table=path_matchup, traffic_dir="./RealTwinDemand/")
#
#     InflowDf_Calibration, InflowEdgeToCalibrate, N_InflowVariable = generate_inflow(path_net,
#                                                                                     MatchupTable_UserInput,
#                                                                                     TurnDf,
#                                                                                     IDRef)
#
#     (TurnToCalibrate, TurnDf_Calibration,
#      RealSummary_Calibration, N_Variable, N_TurnVariable) = generate_turn_summary(TurnDf,
#                                                                                   MatchupTable_UserInput,
#                                                                                   N_InflowVariable)



##############################################################################
# Copyright (c) 2024--, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""Autonomous Vehicle Simulation in SUMO."""

from pathlib import Path
import shutil
import pyufunc as pf

from ._load_config import load_av_configs
from ._sim_av_util import (create_veh_type_attributes,
                           create_sumo_rou_xml,
                           update_sumo_flow_xml,
                           add_veh_types_to_rou,
                           create_sumo_config,
                           run_sumo_simulation)
from ._generate_loop_detector import generate_sumo_loop_detector_add_xml


def prepare_av_configs(dest_dir: str = "") -> bool:
    """Generate configuration files for autonomous vehicles simulation.

    Args:
        dest_dir (str): Directory where the configuration files will be saved.
        If the directory does not exist, use the current working directory.

    Example:
        >>> from realtwin import prepare_av_configs
        >>> prepare_av_configs(dest_dir="path/to/directory")
        >>> "Autonomous vehicle configuration file created at: path/to/directory/AVConfig.yml"
        >>> True

    Returns:
        bool: True if the configuration files were successfully created, False otherwise.
    """

    # check if output directory exists
    if not Path(dest_dir).exists():
        dest_dir = Path.cwd()

    # copy the default configuration file to the output directory
    default_config = Path(__file__).parent.parent / 'data_lib/config_av.yaml'
    fname = Path(dest_dir) / 'config_av.yaml'
    shutil.copy(default_config, fname)
    print(f"  :Autonomous vehicle configuration file created at: {pf.path2linux(fname)}")
    return True


class SimAV:
    """Autonomous Vehicle Simulation

    See Also:
        "config_file": you can run `prepare_av_configs` to generate a default configuration file.
        for details, refer to the `prepare_av_configs` function in the `realtwin` module.
    """

    def __init__(self, path_config: str | Path = "", verbose: bool = True):
        """Initialize the SimAV class.

        Args:
            path_config (str | Path): Path to the configuration file for autonomous vehicle simulation.
            verbose (bool): If True, print detailed information during the simulation.

        See Also:
            "config_file": you can run `prepare_av_configs` to generate a default configuration file.
        """

        # TDD
        if isinstance(path_config, (str, Path)):
            self.path_config = Path(path_config)
        else:
            raise ValueError("  :Error: path_config must be a string or Path object.")

        if not self.path_config.exists():
            raise FileNotFoundError(f"Configuration file not found: {self.path_config}")

        # self.av_config = load_av_configs(self.path_config)
        self.verbose = verbose

    def prepare_av_configs(self, dest_dir: str = "") -> bool:
        """Generate configuration files for autonomous vehicles simulation.

        Args:
            dest_dir (str): Directory where the configuration files will be saved.
            If the directory does not exist, use the current working directory.

        Example:
            >>> from realtwin import prepare_av_configs
            >>> prepare_av_configs(dest_dir="path/to/directory")
            >>> "Autonomous vehicle configuration file created at: path/to/directory/AVConfig.yml"
            >>> True

        Returns:
            bool: True if the configuration files were successfully created, False otherwise.
        """

        return prepare_av_configs(dest_dir)

    def load_config(self, path_config: str) -> bool:
        """Load the configuration file for autonomous vehicles simulation.

        Args:
            path_config (str): Path to the configuration file.

        Returns:
            dict: Loaded configuration data.
        """
        try:
            self.av_config = load_av_configs(path_config)
            return True
        except Exception as e:
            print(f"  :Error loading configuration data: {e}")
        return False

    def run_simulation(self, av_config: str | Path | dict = None) -> bool:
        """Run the autonomous vehicle simulation."""

        # check if av_config is provided
        if isinstance(av_config, (str, Path)):
            # load configuration from file
            self.av_config = load_av_configs(av_config)
        elif isinstance(av_config, dict):
            # use the provided configuration dictionary
            self.av_config = av_config

        if not hasattr(self, 'av_config'):
            try:
                self.av_config = load_av_configs(self.path_config)
            except Exception as e:
                raise Exception(
                    "  :Error: No AV configuration provided or loaded."
                    " Please run `load_config` or provide a valid config file or dictionary."
                ) from e

        # check if required inputs are valid
        if not check_inputs_from_config(self.av_config):
            raise Exception("  :Error: Invalid inputs in the configuration file,"
                            " please check the input directory and files.")

        # get paths from the configuration
        input_dir = self.av_config['input']['input_dir']
        path_net = Path(input_dir) / self.av_config['input']['net_file']
        path_rou = Path(f"{name_without_suffixes(path_net)}.rou.xml")
        path_flow = Path(input_dir) / self.av_config['input']['flow_file']
        path_turn = Path(input_dir) / self.av_config['input']['turn_file']

        # create/update output directory
        output_dir = Path(input_dir) / 'output_AV'
        if output_dir.exists():
            shutil.rmtree(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        # copy input files to output directory
        path_net_output = output_dir / path_net.name
        path_rou_output = output_dir / path_rou.name
        path_flow_output = output_dir / path_flow.name
        path_turn_output = output_dir / path_turn.name
        shutil.copy(path_net, path_net_output)
        # shutil.copy(path_rou, path_rou_output)
        shutil.copy(path_flow, path_flow_output)
        shutil.copy(path_turn, path_turn_output)

        # generate loop detector add XML file
        path_add = output_dir / 'detector.add.xml'
        if not path_add.exists():
            generate_sumo_loop_detector_add_xml(path_net_output,
                                                detector_type="E1",
                                                add_fname=path_add.name,
                                                dest_dir=output_dir)

        # create vehicle type attributes
        veh_types = self.av_config['veh_types']
        veh_type_attributes = create_veh_type_attributes(self.av_config)

        # update SUMO flow XML with new vehicle types
        update_sumo_flow_xml(path_flow_output, veh_types, veh_type_attributes)

        # create SUMO route XML
        create_sumo_rou_xml(path_net_output, path_flow_output, path_turn_output, path_rou_output)

        # add vehicle types to the route file
        add_veh_types_to_rou(path_rou_output, veh_types, veh_type_attributes)

        # create SUMO configuration file
        sim_name = self.av_config.get('sim_name', 'chatt')
        sim_start = self.av_config.get('sim_start', 0)
        sim_time = self.av_config.get('sim_time', 3600)
        sim_end = sim_start + sim_time
        path_cfg = output_dir / f'{sim_name}.sumocfg'
        create_sumo_config(path_cfg,
                           sim_name=sim_name,
                           sim_start=sim_start,
                           sim_end=sim_end)

        # run simulation
        run_sumo_simulation(path_cfg, sim_time=sim_time)
        print(f"  :Simulation completed successfully. Output files are saved in: {output_dir}")
        return True


def check_inputs_from_config(av_config: dict) -> bool:
    """Check if the required inputs are present in the configuration"""

    input_dict = av_config.get('input', {})

    if not input_dict:
        print("  :Warning: No inputs found in the configuration.")
        return False

    input_dir = input_dict.get('input_dir', '')
    if not input_dir:
        print("  :Warning: Input directory not specified in the configuration.")
        return False
    if not Path(input_dir).exists():
        print(f"  :Warning: Input directory does not exist: {input_dir}")
        return False
    if not Path(input_dir).is_dir():
        print(f"  :Warning: Input path is not a directory: {input_dir}")
        return False
    print(f"  :Input directory is valid: {input_dir}")

    # check net, flow, rou and turn files
    net_file = input_dict.get('net_file', '')
    if not net_file:
        print("  :Warning: Network file not specified in the configuration.")
        return False
    path_net = Path(input_dir) / net_file
    if not path_net.exists():
        print(f"  :Warning: Network file does not exist: {path_net}")
        return False

    flow_file = input_dict.get('flow_file', '')
    if not flow_file:
        print("  :Warning: Flow file not specified in the configuration.")
        return False
    path_flow = Path(input_dir) / flow_file
    if not path_flow.exists():
        print(f"  :Warning: Flow file does not exist: {path_flow}")
        return False

    turn_file = input_dict.get('turn_file', '')
    if not turn_file:
        print("  :Warning: Turn file not specified in the configuration.")
        return False
    path_turn = Path(input_dir) / turn_file
    if not path_turn.exists():
        print(f"  :Warning: Turn file does not exist: {path_turn}")
        return False
    print("  :All input files are valid.")
    return True


def name_without_suffixes(p: Path) -> str:
    """Return the name of the file without any suffixes."""
    # peel off each suffix until none remain
    while p.suffix:
        p = p.with_suffix('')
    return p.name



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

# Uncomment this line if you want to use the sumo calibration function from the previous code snippet
# from .calibration_sumo import cali_sumo

# Updated import with third-party library
from .calibration_sumo import cali_sumo
from .calibration_aimsun import cali_aimsun
from .calibration_vissim import cali_vissim

from .algo_sumo.cali_behavior import BehaviorCali
from .algo_sumo.cali_turn_inflow import TurnInflowCali

__all__ = [
    # Calibration functions for different simulators
    "cali_sumo",
    "cali_aimsun",
    "cali_vissim",

    # Calibration algorithms for SUMO simulator
    "BehaviorCali",
    "TurnInflowCali",
]


##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################


import os
import yaml
import re
from pathlib import Path
from zipfile import ZipFile

import pyufunc as pf
from rich.console import Console
console = Console()

from realtwin.util_lib.check_lon_lat_from_list_of_coords import detect_coord_order
from realtwin.util_lib.get_bbox_from_list_of_coords import get_bbox_from_vertices


def load_input_configs(path_config: str) -> dict:
    """load input configuration from yaml file

    Args:
        path_config (str): the path of the configuration file in yaml format

    Raises:
        FileNotFoundError: if the file is not found
        ValueError: if the file is not in yaml format

    Returns:
        dict: the dictionary of the configuration data
    """

    if isinstance(path_config, Path):
        path_config = str(path_config)

    # TDD check whether the file exists and is a yaml file
    if not os.path.exists(path_config):
        raise FileNotFoundError(f"  :File not found: {path_config}")

    if not (path_config.endswith('.yaml') or path_config.endswith('.yml')):
        raise ValueError(f"  :File is not in yaml format: {path_config}")

    # read the yaml file and return the configuration dictionary
    with open(path_config, 'r', encoding="utf-8") as yaml_data:
        config = yaml.safe_load(yaml_data)

    # check whether input_dir exists
    if config.get('input_dir') is None:
        # set input_dir to current working directory if not specified
        config['input_dir'] = pf.path2linux(os.getcwd())
    else:

        # check if the input_dir is a valid path
        if not Path(config['input_dir']).exists():
            config['input_dir'] = pf.path2linux(os.getcwd())

        # convert input_dir to linux format
        config['input_dir'] = pf.path2linux(config['input_dir'])

    # check output_dir from input configuration file
    if config.get('output_dir') is None:
        # set output_dir to input_dir/output if not specified
        config['output_dir'] = pf.path2linux(os.path.join(config['input_dir'], 'output'))
    elif not os.path.exists(config['output_dir']):
        config['output_dir'] = pf.path2linux(os.path.join(config['input_dir'], 'output'))

    # check whether key sections exist in the configuration file
    key_sections = ["Traffic", 'Network', 'Control']
    for key in key_sections:
        if key not in config:
            console.log(f"[bold]{key} section is not found in the configuration file.")

    # update network bbox if vertices are provided in the input configuration file
    if vertices := config.get('Network', {}).get('NetworkVertices'):

        # update vertices to a list of tuples in (lon, lat) format
        if isinstance(vertices, str):
            # Regular expression to extract the coordinate pairs
            pattern = r"\((-?\d+\.\d+),\s*(-?\d+\.\d+)\)"
            matches = re.findall(pattern, vertices)
            vertices = [(float(match[0]), float(match[1])) for match in matches]
        elif isinstance(vertices, (list, tuple)):
            # Check if the list contains tuples
            if all(isinstance(item, (list, tuple)) and len(item) == 2 for item in vertices):
                vertices = [(float(item[0]), float(item[1])) for item in vertices]
            else:
                raise ValueError("Invalid format: List must contain list/tuple of [lon, lat] or (lat, lon) pairs.")
        else:
            raise ValueError("Invalid format: vertices must be a string or list/tuple.")

        # detect the coordinate order
        coord_order = detect_coord_order(vertices)
        if coord_order == "lat, lon":
            vertices = [[float(lon), float(lat)] for (lat, lon) in vertices]
        elif coord_order == "ambiguous":
            console.log("  :Coordinate order is ambiguous. Assuming (lon, lat) format.")

        # update the vertices in the configuration file
        config['Network']['NetworkVertices'] = vertices

        # update the bounding box if it is not provided
        bbox = config.get('Network', {}).get('Net_BBox')
        if not bbox:
            config['Network']['Net_BBox'] = get_bbox_from_vertices(vertices)

    # check whether demo mode is enabled
    config = check_demo_data_enabled(config)

    return config


def check_demo_data_enabled(config: dict) -> dict:
    """Check if demo data is enabled in the configuration.

    Args:
        config (dict): The configuration dictionary.

    Returns:
        dict: The configuration dictionary with demo data check.
    """
    # check whether demo mode is enabled
    AVAILABLE_DEMO_DATA = ["chattanooga"]
    if demo_data := config.get('demo_data'):
        if not isinstance(demo_data, str):
            config['demo_data'] = None
            console.log("  :Demo data is not a string. Demo mode is disabled.")

        elif demo_data.lower() in AVAILABLE_DEMO_DATA:
            try:
                # copy demo data to the input directory
                demo_data_path = pf.path2linux(
                    Path(__file__).parent.parent.parent / "data_lib" / f"{demo_data.lower()}.zip")

                with ZipFile(demo_data_path, 'r') as zip_ref:
                    extract_path = os.path.splitext(demo_data_path)[0]
                    os.makedirs(extract_path, exist_ok=True)
                    zip_ref.extractall(config['input_dir'])
                console.log(
                    f"  :Demo data {demo_data} extracted to {config['input_dir']}.")
                # update input directory to the extracted demo data
                config["input_dir"] = pf.path2linux(
                    Path(config['input_dir']) / f"{demo_data.lower()}")
                config["Network"]["NetworkName"] = demo_data
                # use dummy coordinates to make sure program works (in generate_inputs)
                config["Network"]["NetworkVertices"] = [[-85.14977588011192, 35.040346288414916],
                                                        [-85.15823020212477,
                                                            35.04345144844759],
                                                        [-85.15829457513502,
                                                            35.043293338482925],
                                                        [-85.14986171079225, 35.04018378032611]]
            except Exception as e:
                console.log(
                    f"  :Demo data {demo_data} extraction failed for {e}. Demo mode is disabled.")
                config['demo_data'] = None
        else:
            config['demo_data'] = None
            console.log(
                f"Demo data {demo_data} currently not available. Available demo data: {AVAILABLE_DEMO_DATA}")
            console.log("Demo mode is disabled.")

    return config



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host Control element of Abstract scenario
'''


class Control:
    """Control class to host the signal element of the abstract scenario.
    """
    def __init__(self):
        """Initialize the Control class with the signal element as an empty dictionary."""
        self.Signal = {}

    def isEmpty(self):
        """Check if the Control element is empty."""
        return not bool(self.Signal)



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Control of module imports for the RealTwin package. """


from ._load_config import load_av_configs
from .sim_av import SimAV, prepare_av_configs
from ._generate_loop_detector import generate_sumo_loop_detector_add_xml


__all__ = [
    'SimAV',
    'prepare_av_configs',
    'load_av_configs',
    'generate_sumo_loop_detector_add_xml',
]



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host a unique AbstractScenario definition
'''

# import four elements of AbstractScenario
from ._supply import Supply
from ._demand import Demand
from ._behavior import Behavior
from ._route import Route
from ._trafficControl import TrafficControl


class ConcreteScenario:
    """Initialize and Generate Concrete Scenario from Abstract Scenario"""
    def __init__(self):
        self.Supply = Supply()
        self.Demand = Demand()
        self.Behavior = Behavior()
        self.Route = Route()
        self.TrafficControl = TrafficControl()

    def is_empty(self):
        """Check if the ConcreteScenario object is empty."""
        pass

    def get_unified_scenario(self, AbsScn):
        """Generate Concrete Scenario from Abstract Scenario"""
        # copy the input_config from AbstractScenario incase it is needed
        self.input_config = AbsScn.input_config

        # generate concrete scenario
        self.Supply.generate_network(AbsScn)
        self.Demand.generate_traffic(AbsScn)
        self.Behavior.ApplicationInterpreter(AbsScn)
        self.Route.generate_route(AbsScn)
        self.TrafficControl.generate_control(AbsScn)
